#ifndef __DAG_SP_NODES_ONLY__
#define __DAG_SP_NODES_ONLY__

#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "../problem.h"
#include "functionQ.h"


class DAG_SP_NODES_ONLY : public Problem
{
public:
	DAG_SP_NODES_ONLY(int _node_num, int _edge_num_max) 
		: node_num(_node_num), edge_num(0), edge_num_max(_edge_num_max), last_node(0), f(NULL), 
		reduce_node_map(NULL),
		edge_start(node_num)
	{
		nodes = new Node[node_num+2];
		edges = new Edge[edge_num_max];
		D = (new double[node_num+2])+1;
		back_ptr = (new int[node_num+2])+1;

		nodes[0].first_in = 0;

		time_ptr = NULL;
	}
	~DAG_SP_NODES_ONLY()
	{
		if (f) delete f;
		delete [] nodes;
		delete [] edges;
		delete [] (D-1);
		delete [] (back_ptr-1);
	}
	void SetEdgeStart(int _edge_start) { edge_start = _edge_start; }
	int GetEdgeStart() { return edge_start; }


	void SetChain(int N, int K)
	{
		int i, a, b;
		for (a=0; a<K; a++)
		{
			AddEdge(-1, a);
		}
		for (i=1; i<N; i++)
		for (b=0; b<K; b++)
		for (a=0; a<K; a++)
		{
			AddEdge((i-1)*K+a, i*K+b);
		}
		for (a=0; a<K; a++)
		{
			AddEdge((N-1)*K+a, N*K);
		}
		FinalizeDAG();
	}
	void SetRandomFunction(long unsigned int seed)
	{
		srand(seed);
		f = new FunctionQ(node_num);
		f->SetRandom(seed);
	}

	void AddEdge(int i, int j) // j should be monotonic between consecutive calls to AddEdge()
	{
		//printf("%d %d [%f]\n", i, j, weight);
		if (i<-1 || i>=j || j<last_node-1 || j>node_num)  { printf("DAG_SP_NODES_ONLY::AddEdge(): invalid call\n"); exit(1); }
		if (edge_num >= edge_num_max) { printf("DAG_SP_NODES_ONLY::AddEdge(): too many edges\n"); exit(1); }
		while (j >= last_node)
		{
			last_node ++;
			nodes[last_node].first_in = nodes[last_node-1].first_in;
		}
		edges[edge_num].from = i;
		nodes[last_node].first_in = ++ edge_num;
	}

	void FinalizeDAG()
	{
		while (node_num >= last_node)
		{
			last_node ++;
			nodes[last_node].first_in = nodes[last_node-1].first_in;
		}
	}

	int get_n() override
	{
		return node_num+edge_num;
	} 

//	int get_n_nonzero() override
//	{
//		return node_num;
//	} 

	int get_N() override
	{
		return node_num+edge_num;
	}

	void init_grad(Vector g) override
	{
		memset(g+edge_start, 0, edge_num*sizeof(double));
	}
		
	void destroy_atom(Atom a) override
	{
		free(a);
	}

	Atom get_init_atom()
	{
		int i;
		double* g = new double[node_num+edge_num];
		for (i=0; i<node_num+edge_num; i++) g[i] = (rand() % 100000);
		init_grad(g);
		Atom a = min_oracle(g, NO_TIME_LIMIT);
		delete [] g;
		return a;
	}

	Atom get_lowest_atom()
	{
		double* g = new double[node_num+edge_num];
		memset(g, 0, node_num*sizeof(double));
		init_grad(g);
		Atom a = min_oracle(g, NO_TIME_LIMIT);
		delete [] g;
		return a;
	}

	void atom_to_extended_vector(Atom _a, ExtendedVector x) override
	{
		ATOM* a = (ATOM*)_a;
		int i;
		memset(x, 0, node_num*sizeof(double));
		memset(x+edge_start, 0, edge_num*sizeof(double));
		for (i=0; i<a->len; i++)
		{
			int e = a->e_list[i];
			x[edge_start+e] = 1.0;
			if (i > 0) x[edges[e].from] = 1.0;
		}
	}

	void add_atom_to_extended_vector(double gamma, Atom _a, ExtendedVector x) override
	{
		ATOM* a = (ATOM*)_a;
		int i;
		for (i=0; i<a->len; i++)
		{
			int e = a->e_list[i];
			x[edge_start+e] += gamma;
			if (i > 0) x[edges[e].from] += gamma;
		}
	}

	double dot_product(Atom _a, Vector x) override
	{
		ATOM* a = (ATOM*)_a;
		int i;
		double sum = 0;

		if (a->len>0) sum += x[edge_start+a->e_list[0]];
		for (i=1; i<a->len; i++)
		{
			int e = a->e_list[i];
			sum += x[edge_start+e];
			sum += x[edges[e].from];
		}

		return sum;
	}

	int compare_atoms(Atom _a, Atom _b) override
	{
		ATOM* a = (ATOM*) _a;
		ATOM* b = (ATOM*) _b;
		if (a->len < b->len) return -1;
		if (a->len > b->len) return +1;
		return memcmp(a->e_list, b->e_list, a->len*sizeof(int));
	}

	double eval(ExtendedVector x) override
	{
		return f->eval(x);
	}

	void grad(ExtendedVector x, Vector g, double* fx) override
	{
		f->grad(x, g, fx);
	}

	double line_search(ExtendedVector x, ExtendedVector y, double* delta_f, double* delta_f1, double* delta_f_unconstrained) override
	{
		return f->line_search(x, y, delta_f, delta_f1, delta_f_unconstrained);
	}

	Atom min_oracle(Vector g, double time_limit) override
	{
		if (time_ptr) *time_ptr += edge_num;
		int i, j, e, e_opt;
		double cost, cost_opt;

		D[-1] = 0;
		for (j=0; j<=node_num; j++)
		{
			e_opt = -1;
			cost_opt = 1e300;
			double gj = (j == node_num) ? 0 : g[j];
			for (e=nodes[j].first_in; e<nodes[j+1].first_in; e++)
			{
				i = edges[e].from;
				cost = D[i] + gj;
				if (cost_opt > cost)
				{
					e_opt = e;
					cost_opt = cost;
				}
			}
			back_ptr[j] = e_opt;
			D[j] = cost_opt;
		}
		return trace_back_ptr(back_ptr);
	}

	Atom minmax_oracle(Vector g, Vector x, Atom& v, double time_limit) override
	{
		if (time_ptr) *time_ptr += edge_num;
		int i, j, e;

		double* Dmax = (new double[node_num+2]) + 1;
		int* back_ptr_max = new int[node_num+1];
		double* Dmin = D;
		int* back_ptr_min = back_ptr;

		Dmax[-1] = Dmin[-1] = 0;

		for (j=0; j<=node_num; j++)
		{
			int e_opt_min = -1;
			int e_opt_max = -1;
			double cost_opt_min = 1e300;
			double cost_opt_max = -1e300;
			double gj = (j == node_num) ? 0 : g[j];
			for (e=nodes[j].first_in; e<nodes[j+1].first_in; e++)
			{
				i = edges[e].from;
				if ( 1 )
				{
					double cost = Dmin[i] + gj;
					if (cost_opt_min > cost)
					{
						e_opt_min = e;
						cost_opt_min = cost;
					}
				}
				if (x[edge_start+e] > 0)
				{
					if (x[edge_start+e] < 1e-10)
					{
						delete [] (Dmax-1);
						delete [] back_ptr_max;
						x[edge_start+e] = 0;
						adjust_x(x);
						return minmax_oracle(g, x, v, time_limit);
					}
					double cost = Dmax[i] + gj;
					if (cost_opt_max < cost)
					{
						e_opt_max = e;
						cost_opt_max = cost;
					}
				}
			}
			back_ptr_min[j] = e_opt_min;
			back_ptr_max[j] = e_opt_max;
			Dmin[j] = cost_opt_min;
			Dmax[j] = cost_opt_max;
		}
		Atom s = trace_back_ptr(back_ptr_min);
		v = trace_back_ptr(back_ptr_max);
		delete [] (Dmax-1);
		delete [] back_ptr_max;

		return s;
	}

	double get_alpha(Vector x, Atom _v)
	{
		ATOM* v = (ATOM*)_v;
		int k, e;
		double alpha = 1;
		for (k=0; k<v->len; k++)
		{
			e = v->e_list[k];
			if (alpha > x[edge_start+e]) alpha = x[edge_start+e];
		}
		return alpha;
	}

	double move_DICG(Vector x, Atom _s, Atom _v) override
	{
		double alpha = get_alpha(x, _v);
		move_DICG(x, _s, _v, alpha);
		return alpha;
	}

	void move_DICG(Vector x, Atom _s, Atom _v, double alpha)
	{
		ATOM* s = (ATOM*)_s;
		ATOM* v = (ATOM*)_v;
		int d, i, k, e;

		for (d=0; d<2; d++)
		{
			ATOM* a = (d==0) ? v : s;
			for (k=0; k<a->len; k++)
			{
				e = a->e_list[k];
				i = edges[e].from;
				if (i >= 0) x[i] -= alpha;
				x[edge_start+e] -= alpha;
			}
			alpha = -alpha;
		}
	}

	Problem* reduce(int* edge_list, int e_num)
	{
		int i, j, e, k;

		memset(back_ptr-1, 0, (node_num+2)*sizeof(int));
		for (k=j=0; k<e_num; k++)
		{
			e = edge_list[k]; i = edges[e].from; while (nodes[j+1].first_in <= e) j ++;
			back_ptr[i] |= (back_ptr[i] & 1) ? 2 : 1;
			back_ptr[j] |= (back_ptr[j] & 4) ? 8 : 4;
		}
		back_ptr[-1] = back_ptr[node_num] = 2+8;

		DAG_SP_NODES_ONLY* Q = new DAG_SP_NODES_ONLY(node_num, e_num);
		int* edge_map = new int[edge_num];
		Q->contract_data.P = this;
		Q->contract_data.edge_list = edge_list;
		Q->contract_data.e_num = e_num;
		Q->contract_data.edge_map = edge_map;

		int node_num_new = 0;
		int* node_map;
		if (reduce_node_map) node_map = reduce_node_map;
		else node_map = (new int[node_num+2])+1;
		node_map[-1] = -1;
		node_map[0] = -2;
		for (k=j=0; k<e_num; k++)
		{
			e = edge_list[k]; i = edges[e].from; while (nodes[j+1].first_in <= e) node_map[++ j] = -2;
			if (node_map[i] == -2) node_map[i] = node_num_new ++;
			if ((back_ptr[i] & 2) == 0 && (back_ptr[j] & 8) == 0) // outdegree(i) = indegree(j) = 1    ;   contract e
			{
				if (node_map[j] != -2) { printf("Error in reduce\n"); exit(1); }
				node_map[j] = node_map[i];
				edge_map[k] = node_map[i];
			}
			else
			{
				if (node_map[j] == -2) node_map[j] = node_num_new ++;
				edge_map[k] = -2-Q->edge_num;
				Q->AddEdge(node_map[i], node_map[j]);
			}
		}

		Q->node_num = node_num_new-1;
		Q->FinalizeDAG();

		if (f) Q->f = f->shrink(Q->node_num, node_map);
		if (!reduce_node_map) delete [] (node_map-1);
			
		return Q;
	}
	
	Problem* reduce(Vector g, ExtendedVector x, Atom _s, ReducePolicy& policy, double time_limit) override
	{

		if (time_ptr) *time_ptr += edge_num;
		int i, j, e, e_num = 0;
		int* edge_list = new int[edge_num];

		ATOM* s = (ATOM*) _s;
		int* s_ptr = s->e_list;
		int* s_end = s_ptr+s->len;
		if (policy.is_minimal())
		{
			for (e=0; e<edge_num; e++)
			{
				while (s_ptr < s_end && *s_ptr<e) s_ptr ++;

				if (x[edge_start+e] > 0 || (s_ptr < s_end && *s_ptr==e))
				{
					edge_list[e_num ++] = e;
				}
			}
		}
		else
		{
			double* Drev = (new double[node_num+2 + edge_num]) + 1;
			double* min_marginals = Drev + node_num + 1;
			for (i=-1; i<node_num; i++) Drev[i] = 1e300;
			Drev[node_num] = 0;

			for (j=node_num; j>=0; j--)
			{
				for (e=nodes[j].first_in; e<nodes[j+1].first_in; e++)
				{
					i = edges[e].from;
					double cost = Drev[j] + ((i >= 0) ? g[i] : 0);
					if (Drev[i] > cost) Drev[i] = cost;
				}
			}

			int min_marginal_num = 0;
			for (j=0; j<=node_num; j++)
			{
				for (e=nodes[j].first_in; e<nodes[j+1].first_in; e++)
				{
					while (s_ptr < s_end && *s_ptr<e) s_ptr ++;
					if (x[edge_start+e] > 0 || (s_ptr < s_end && *s_ptr==e))
					{
						edge_list[e] = 2;
					}
					else
					{
						i = edges[e].from;
						min_marginals[min_marginal_num ++] = D[i] + Drev[j];
						edge_list[e] = 0;
					}
				}
			}
			int m_min = edge_num - min_marginal_num;
			int m = policy.get_size(edge_num, m_min) - m_min - 1;
			double cost_threshold = -1e300;
			if (m >= 0)
			{
				std::nth_element(min_marginals, min_marginals+m, min_marginals+min_marginal_num);
				cost_threshold = min_marginals[m];
			}

			memset(back_ptr-1, 0, (node_num+2)*sizeof(int));
			back_ptr[-1]       |= 1;
			back_ptr[node_num] |= 2;
			for (j=node_num; j>=0; j--)
			{
				if (!(back_ptr[j] & 2)) continue;
				for (e=nodes[j].first_in; e<nodes[j+1].first_in; e++)
				{
					i = edges[e].from;
					if (edge_list[e] == 0 && D[i] + Drev[j] < cost_threshold) edge_list[e] = 1;
					if (edge_list[e] > 0) back_ptr[i] |= 2;
				}
			}
			for (j=0; j<=node_num; j++)
			{
				for (e=nodes[j].first_in; e<nodes[j+1].first_in; e++)
				{
					if (edge_list[e] == 0) continue;
					if (edge_list[e] == 1)
					{
						i = edges[e].from;
						if (!(back_ptr[i] & 1)) continue;
					}
					back_ptr[j] |= 1;
					edge_list[e_num ++] = e;
				}
			}

			delete [] (Drev-1);
		}

		Problem* Q = reduce(edge_list, e_num);
		return Q;
	}

	Atom shrink_atom(Problem* _B, Atom _a)
	{
		ATOM* a = (ATOM*)_a;
		DAG_SP_NODES_ONLY* B = (DAG_SP_NODES_ONLY*)_B;

		int* edge_list = B->contract_data.edge_list;
		int k, e_num = B->contract_data.e_num;
		int* edge_map = B->contract_data.edge_map;

		int* a_ptr = a->e_list;
		int* a_end = a_ptr + a->len;

		a->len = 0;
		for (k=0; a_ptr < a_end; a_ptr ++)
		{
			while (k < e_num && edge_list[k] < *a_ptr) k ++;
			if (k == e_num || edge_list[k] != *a_ptr) { destroy_atom(a); return NULL; } // { printf("Error in shrink_atom()\n"); exit(1); }
			if (edge_map[k] <= -2)
			{
				a->e_list[a->len ++] = -2-edge_map[k];
			}
		}
		return a;
	}
	Atom unshrink_atom(Problem* _A, Atom _a)
	{
		ATOM* a = (ATOM*)_a;
		DAG_SP_NODES_ONLY* A = (DAG_SP_NODES_ONLY*)_A;

		int* edge_list = A->contract_data.edge_list;
		int k, e_num = A->contract_data.e_num;
		int* edge_map = A->contract_data.edge_map;

		int* a_ptr = a->e_list;
		int* a_end = a_ptr + a->len;

		memset(A->back_ptr-1, 0, (A->node_num+2)*sizeof(int));

		for (k=0; k<a->len; k++)
		{
			int e = a->e_list[k];
			A->back_ptr[A->edges[e].from] = 1;
		}

		int* new_list = back_ptr-1;
		int new_len = 0;

		for (k=0; k<e_num; k++)
		{
			if (edge_map[k] == -1) continue;
			if (edge_map[k] >= 0)
			{
				if (!A->back_ptr[edge_map[k]]) continue;
			}
			else
			{
				int e = -2-edge_map[k];
				while (a_ptr < a_end && *a_ptr < e) a_ptr ++;
				if (a_ptr == a_end || *a_ptr != e) continue;
			}
			new_list[new_len++] = edge_list[k];
		}
		free(_a);
		a = (ATOM*) malloc(sizeof(ATOM)+(new_len-1)*sizeof(int));
		a->len = new_len;
		memcpy(a->e_list, new_list, new_len*sizeof(int));
		return a;
	}
	Atom reshape_atom(Problem* A, Problem* B, Atom a) override
	{
		if (A == this) A = NULL;
		if (B == this) B = NULL;
		if (A) a = unshrink_atom(A, a);
		if (B) a = shrink_atom(B, a);
		return a;
	}
	void reshape_x(Problem* _A, Problem* _B, ExtendedVector x) override
	{
		if (_A == this) _A = NULL;
		if (_B == this) _B = NULL;
		if (_A)
		{
			DAG_SP_NODES_ONLY* A = (DAG_SP_NODES_ONLY*)_A;
			double* y = new double[A->node_num+A->edge_num];
			memcpy(y, x, (A->node_num+A->edge_num)*sizeof(double));
			unshrink_x(_A, y, y+A->edge_start, x, x+edge_start);
			delete [] y;
		}
		if (_B) shrink_x(_B, x, x+edge_start, x, x+((DAG_SP_NODES_ONLY*)_B)->edge_start);
	}
	void shrink_x(Problem* _B, double* x, double* xe, double* b, double* be)
	{
		DAG_SP_NODES_ONLY* B = (DAG_SP_NODES_ONLY*)_B;

		int* edge_list = B->contract_data.edge_list;
		int k, e_num = B->contract_data.e_num;
		int* edge_map = B->contract_data.edge_map;

		memset(b, 0, B->node_num*sizeof(double));

		for (k=0; k<e_num; k++)
		{
			if (edge_map[k] <= -2)
			{
				int e = -2-edge_map[k];
				double z = be[e] = xe[edge_list[k]];
				int i = B->edges[e].from;
				if (i>=0) b[i] += z;
			}
		}
	}
	void unshrink_x(Problem* _B, double* b, double* be, double* x, double* xe)
	{
		DAG_SP_NODES_ONLY* B = (DAG_SP_NODES_ONLY*)_B;

		int* edge_list = B->contract_data.edge_list;
		int k, e_num = B->contract_data.e_num;
		int* edge_map = B->contract_data.edge_map;

		memset(x, 0, node_num*sizeof(double));
		memset(xe, 0, edge_num*sizeof(double));

		for (k=0; k<e_num; k++)
		{
			if (edge_map[k] == -1) continue;
			double z;
			if (edge_map[k] <= -2)
			{
				int e = -2-edge_map[k];
				z = be[e];
			}
			else if (edge_map[k] >= 0)
			{
				z = b[edge_map[k]];
			}
			int e = edge_list[k];
			xe[e] = z;
			int i = edges[e].from;
			if (i>=0) x[i] += z;
		}
	}

	void adjust_x(ExtendedVector x, double eps=0)
	{
		int i, e;
		int* first_out = (new int[node_num+edge_num+1])+1;
		int* next_out = first_out+node_num;
		for (i=-1; i<node_num; i++) first_out[i] = -1;
		for (e=0; e<edge_num; e++)
		{
			i = edges[e].from;
			next_out[e] = first_out[i];
			first_out[i] = e;
		}

		for (i=-1; i<node_num; i++)
		{
			double L = 0, R = 0;
			if (i == -1) L = 1;
			else
			{
				for (e=nodes[i].first_in; e<nodes[i+1].first_in; e++) L += x[edge_start+e];
			}
			for (e=first_out[i]; e>=0; e=next_out[e])
			{
				if (x[edge_start+e] < eps) x[edge_start+e] = 0;
				R += x[edge_start+e];
			}
			if (R <= 0)
			{
				e = first_out[i];
				if (e >= 0) x[edge_start+e] = L;
			}
			else
			{
				L /= R;
				for (e=first_out[i]; e>=0; e=next_out[e])
				{
					x[edge_start+e] *= L;
				}
			}
		}

		delete [] (first_out-1);

		memset(x, 0, node_num*sizeof(double));

		for (e=0; e<edge_num; e++)
		{
			i = edges[e].from;
			if (i >= 0) x[i] += x[edge_start+e];
		}
	}

	void PrintAtom(Atom _a) override
	{
		ATOM* a = (ATOM*)_a;
		int i;
		for (i=0; i<a->len; i++) printf("%d ", a->e_list[i]);
	}

	int get_node_num() { return node_num; }
	int get_edge_num() { return edge_num; }

private:
	struct ATOM
	{
		int len;
		int e_list[1];
	};
	typedef int NodeId;
	typedef int EdgeId;
	struct Node
	{
		EdgeId first_in;
	};
	struct Edge
	{
		NodeId from;
	};
	Node* nodes;
	Edge* edges;
	int last_node, node_num; 
	int edge_num, edge_num_max;
	int edge_start;
	FunctionQ* f;

	struct ContractData
	{
		ContractData() : P(NULL), edge_list(NULL), e_num(0), edge_map(NULL) {}
		~ContractData()
		{
			if (edge_list) delete [] edge_list;
			if (edge_map) delete [] edge_map;
		}
		Problem* P; // parent
		int e_num;
		int* edge_list; // of size e_num
		int* edge_map; // of size e_num
	} contract_data;

	// for temporary computation
	double* D; // of size last_node
	int* back_ptr; // of size last_node

	Atom trace_back_ptr(int* back_ptr)
	{
		int i, e, len = 0;
		for (e=back_ptr[node_num]; ; )
		{
			if (e < 0) { printf("Error: disconnected graph!\n"); exit(1); }
			len ++;
			i = edges[e].from;
			if (i < 0) break;
			e = back_ptr[i];
		}
		ATOM* a = (ATOM*) malloc(sizeof(ATOM)+(len-1)*sizeof(int));
		a->len = len;
		for (e=back_ptr[node_num]; ; )
		{
			if (e < 0) { printf("Error: disconnected graph!\n"); exit(1); }
			a->e_list[-- len] = e;
			if (len == 0) break;
			i = edges[e].from;
			e = back_ptr[i];
		}
		return a;
	}

public:
	int* reduce_node_map;
};


#endif
