#ifndef __PM__
#define __PM__

#include <string.h>
#include <stdlib.h>
#include "../problem.h"
#include "function.h"
#include "blossom5/PerfectMatching.h"

#include <algorithm>
#include <vector>

class PM : public Problem
{
public:
	PM(const char* filename, bool add_default_matching=true) : edge_num(0), edge_num_max(0),
		shrink_mapping(NULL), f(NULL), P(NULL), g_prev(NULL),
		warm_start(false), f_method(0)
	{
		FILE* fp = fopen(filename, "r");
		if (!fp) { printf("cannot open %s\n", filename); exit(1); }
		if (fscanf(fp, "%d", &node_num) != 1 || node_num < 2 || (node_num & 1)) { printf("%s: bad format\n", filename); exit(1); }
		int i, j;
		edge_num_max = node_num;
		edges = new int[2*edge_num_max];
		if (add_default_matching)
		{
			for (i=0; i<node_num; i+=2) AddEdge(i, i+1);
		}
		while (fscanf(fp, "%d %d", &i, &j) == 2)
		{
			if (edge_num >= edge_num_max)
			{
				int* edges_old = edges;
				edge_num_max *= 2;
				edges = new int[2*edge_num_max];
				memcpy(edges, edges_old, 2*edge_num*sizeof(int));
				delete [] edges_old;
			}
			AddEdge(i, j);
		}
		fclose(fp);
	}
	PM(int _node_num, int _edge_num_max) : node_num(_node_num), edge_num(0), edge_num_max(_edge_num_max), 
		shrink_mapping(NULL), f(NULL), P(NULL), g_prev(NULL),
		warm_start(false), f_method(0)
	{
		edges = new int[2*edge_num_max];
	}
	//PM(int N) : node_num(2*N), edge_num(0), edge_num_max(N*N),
	//	shrink_mapping(NULL), f(NULL), P(NULL), g_prev(NULL),
	//	warm_start(false), f_method(0)
	//{
	//	int i, j;
	//	edges = new int[2*edge_num_max];
	//	for (i=0; i<N; i++)
	//	for (j=0; j<N; j++)
	//	{
	//		AddEdge(i, N+j);
	//	}
	//}
	void SetBirkhoff()
	{
		int N = node_num / 2;
		int i, j;
		for (i=0; i<N; i++)
		for (j=0; j<N; j++)
		{
			AddEdge(i, N+j);
		}
	}
	void SetRegularBipartite(int degree)
	{
		int i, N = node_num / 2;
		int* buf = new int[N];
		for ( ; degree > 0; degree --)
		{
			generate_permutation(buf, N);
			for (i=0; i<N; i++) AddEdge(i, N+buf[i]);
		}
		delete [] buf;
	}

	~PM()
	{
		if (f) delete f;
		delete [] edges;
		if (shrink_mapping) delete [] shrink_mapping;
		if (P) delete P;
		if (g_prev) delete [] g_prev;
	}
	bool warm_start;

	int RandomInteger(int N) // returns random integer in [0,N-1]
	{
		while ( 1 )
		{
			int x = (int) ( (((double)N) * rand()) / RAND_MAX );
			if (x<N) return x;
		}
	}
	void generate_permutation(int *buf, int n)
	{
		int i, j;

		for (i=0; i<n; i++) buf[i] = i;
		for (i=0; i<n-1; i++)
		{
			j = i + RandomInteger(n-i);
			int tmp = buf[i]; buf[i] = buf[j]; buf[j] = tmp;
		}
	}


	void SetRandomFunction(bool uniform_A, int m, double sigma, long unsigned int seed, int _f_method=0)
	{
		f_method = _f_method;
		f = new Function(edge_num, m, f_method);
		Atom _x0 = get_lowest_atom();
		double* x0 = new double[f->get_N()];
		atom_to_extended_vector(_x0, x0);
		destroy_atom(_x0);
		f->SetRandom(uniform_A, x0, sigma, seed);
		delete [] x0;
	}

	void AddEdge(int i, int j)
	{
		if (i<0 || i>=node_num || j<0 || j>=node_num || i==j)  { printf("PM::AddEdge(): invalid call\n"); exit(1); }
		if (edge_num >= edge_num_max) { printf("PM::AddEdge(): too many edges\n"); exit(1); }
		edges[2*edge_num] = i;
		edges[2*edge_num+1] = j;
		edge_num ++;
	}

	void Print()
	{
		int e;
		for (e=0; e<edge_num; e++) printf("%d-%d ", edges[2*e], edges[2*e+1]);
		printf("\n");
	}

	int get_n() override
	{
		return edge_num;
	} 

	int get_N() override
	{
		return f->get_N();
	}

	void destroy_atom(Atom a) override
	{
		delete [] (int*)a;
	}

	Atom get_init_atom() override
	{
		int i;
		double* g = new double[edge_num];
		for (i=0; i<edge_num; i++) g[i] = rand() / (RAND_MAX+1.0);
		Atom a = min_oracle(g, NO_TIME_LIMIT);
		delete [] g;
		return a;
	}

	Atom get_lowest_atom()
	{
		double* g = new double[edge_num];
		memset(g, 0, edge_num*sizeof(double));
		Atom a = min_oracle(g, NO_TIME_LIMIT);
		delete [] g;
		return a;
	}

	void atom_to_extended_vector(Atom _a, ExtendedVector x) override
	{
		int* a = (int*)_a;
		int i;
		memset(x, 0, edge_num*sizeof(double));
		for (i=0; i<node_num/2; i++) x[a[i]] = 1.0;

		if (f_method > 0)
		{
			double* y = x + edge_num;
			double** A = f->A;
			double* b = f->b;
			int m = f->m;
			// set y = Ax-b

			int i = a[0], d;
			 
			for (d=0; d<m; d++)
			{
				double sum = -b[d];
				for (i=0; i<node_num/2; i++) sum += A[d][a[i]];
				y[d] = sum;
			}

			//int i, d;
			//for (d=0; d<m; d++)
			//{
			//	double sum = -b[d];
			//	for (i=0; i<n; i++) sum += A[d][i]*x[i];
			//	y[d] = sum;
			//}
		}
	}

	void add_atom_to_extended_vector(double gamma, Atom _a, ExtendedVector x) override
	{
		int* a = (int*)_a;
		int i;
		for (i=0; i<node_num/2; i++) x[a[i]] += gamma;

		if (f_method > 0)
		{
			double* y = x + edge_num;
			double** A = f->A;
			double* b = f->b;
			int m = f->m;
			// set y = Ax-b

			int i = a[0], d;

			for (d=0; d<m; d++)
			{
				double sum = -b[d];
				for (i=0; i<node_num/2; i++) sum += A[d][a[i]];
				y[d] += gamma*sum;
			}
		}
	}

	double dot_product(Atom _a, Vector x) override
	{
		int* a = (int*)_a;
		int i;
		double sum = 0;
		for (i=0; i<node_num/2; i++) sum += x[a[i]];
		return sum;
	}

	int compare_atoms(Atom _a, Atom _b) override
	{
		return memcmp(_a, _b, (node_num/2)*sizeof(int));
	}

	double eval(ExtendedVector x) override
	{
		if (time_ptr) *time_ptr += edge_num;
		return f->eval(x);
	}

	void grad(ExtendedVector x, Vector g, double* fx) override
	{
		if (time_ptr) *time_ptr += edge_num;
		f->grad(x, g, fx);
	}

	double line_search(ExtendedVector x, ExtendedVector y, double* delta_f, double* delta_f1, double* delta_f_unconstrained) override
	{
		return f->line_search(x, y, delta_f, delta_f1, delta_f_unconstrained);
	}

	Atom min_oracle(Vector g, double time_limit) override
	{
		int e;
		if (!warm_start && P)
		{
			delete P;
			P = NULL;
		}
		if (!P)
		{
			P = new PerfectMatching(node_num, edge_num);
			if (g_prev) delete [] g_prev;
			g_prev = new double[edge_num];
			for (e=0; e<edge_num; e++)
			{
				P->AddEdge(edges[2*e], edges[2*e+1], g[e]);
				g_prev[e] = g[e];
			}
		}
		else
		{
			P->StartUpdate();
			for (e=0; e<edge_num; e++)
			{
				P->UpdateCost(e, g[e] - g_prev[e]);
				g_prev[e] = g[e];
			}
			P->FinishUpdate();
		}
		P->options.verbose = 0;
		if (!P->Solve(true, time_limit))
		{
			delete P;
			P = NULL;
			return NULL;
		}

		int* a = new int[node_num/2];

		int len = 0;
		for (e=0; e<edge_num; e++)
		{
			if (P->GetSolution(e)) a[len ++] = e;
		}

		return a;
	}

	Atom minmax_oracle(Vector g, Vector x, Atom& _v, double time_limit) override
	{
		const double eps = 1e-10;
		int e, ee, e_num = 0;
		PerfectMatching* P = new PerfectMatching(node_num, edge_num);
		int* map = new int[edge_num];
		for (e=0; e<edge_num; e++)
		{
			if (x[e] <= eps) continue;
			map[e_num ++] = e;
			P->AddEdge(edges[2*e], edges[2*e+1], -g[e]);
		}
		P->options.verbose = 0;
		if (!P->Solve(true, time_limit))
		{
			delete P;
			delete [] map;
			return NULL;
		}

		int* v = new int[node_num/2];
		_v = v;

		int len = 0;
		for (ee=0; ee<e_num; ee++)
		{
			if (P->GetSolution(ee))
			{
				e = map[ee];
				v[len ++] = e;
			}
		}
		delete P;
		delete [] map;

		Atom s = min_oracle(g, time_limit);
		if (!s) destroy_atom(v);
		return s;
	}

	double move_DICG(Vector x, Atom _s, Atom _v) override
	{
		double alpha = get_alpha(x, _v);
		move_DICG(x, _s, _v, alpha);
		return alpha;
	}

	double get_alpha(Vector x, Atom _v)
	{
		int* v = (int*)_v;
		int k, e;
		double alpha = 1;
		for (k=0; k<node_num/2; k++)
		{
			e = v[k];
			if (alpha > x[e]) alpha = x[e];
		}
		return alpha;
	}

	void move_DICG(Vector x, Atom _s, Atom _v, double alpha)
	{
		int* s = (int*)_s;
		int* v = (int*)_v;
		int dir, k, e;

		for (dir=0; dir<2; dir++)
		{
			int* a = (dir == 0) ? v : s;
			for (k=0; k<node_num/2; k++)
			{
				e = a[k];
				x[e] -= alpha;
			}
			alpha = -alpha;
		}

		if (f_method > 0)
		{
			double* y = x + edge_num;
			double** A = f->A;
			double* b = f->b;
			int m = f->m, d;

			for (dir=0; dir<2; dir++)
			{
				int* a = (dir == 0) ? v : s;
				for (k=0; k<node_num/2; k++)
				{
					e = a[k];
					for (d=0; d<m; d++) y[d] -= alpha*A[d][e];
				}
				alpha = -alpha;
			}
		}
	}

	Problem* reduce(Vector g, ExtendedVector x, Atom _s, ReducePolicy& policy, double time_limit) override
	{
		int* s = (int*)_s;

		int e;
		int* map = new int[edge_num];
		for (e=0; e<edge_num; e++) map[e] = -2;
		for (e=0; e<edge_num; e++) { if (x[e] > 0) map[e] = 0; }
		for (e=0; e<node_num/2; e++) { if (map[s[e]] < 0) map[s[e]] = 0; }

		if (!policy.is_minimal())
		{
			double* slacks = new double[2*edge_num];
			double* slacks_sorted = slacks + edge_num;

			ComputePerfectMatchingSlacks(node_num, edge_num, edges, g, slacks, P);
			int slack_num = 0;
			for (e=0; e<edge_num; e++)
			{
				if (map[e] < 0) slacks_sorted[slack_num ++] = slacks[e];
			}
			int e_min = edge_num - slack_num;
			int m = policy.get_size(edge_num, e_min) - e_min - 1;
			if (m >= 0)
			{
				std::nth_element(slacks_sorted, slacks_sorted+m, slacks_sorted+slack_num);
				double slack_threshold = slacks_sorted[m];
				for (e=0; e<edge_num; e++)
				{
					if (slacks[e] < slack_threshold) map[e] = 0;
				}
			}

			delete [] slacks;
		}

		Problem* Q = reduce(map);
		Q->set_time_ptr(time_ptr);
		return Q;
	}

#define REDUCE_PM_COMPLETE

#ifdef REDUCE_PM_COMPLETE
	Problem* reduce(int* map)
	{
		int i, j, k, e, dir, N=0, E=0;

		struct Node
		{
			int first[2];
			int degree;
			int next;
		};
		struct Edge
		{
			int next[2];
		};

		Node* NODES = new Node[node_num];
		Edge* EDGES = new Edge[edge_num];
		
		for (i=0; i<node_num; i++)
		{
			NODES[i].degree = 0;
			NODES[i].next = -1;
			NODES[i].first[0] = NODES[i].first[1] = -1;
		}
		for (e=0; e<edge_num; e++)
		{
			if (map[e] >= 0)
			{
				i = edges[2*e]; j = edges[2*e+1];
				EDGES[e].next[0] = NODES[i].first[0];
				EDGES[e].next[1] = NODES[j].first[1];
				NODES[i].first[0] = e;
				NODES[j].first[1] = e;
				NODES[i].degree ++;
				NODES[j].degree ++;
				E ++;
			}
		}

		int list_first = -1;
		for (i=0; i<node_num; i++)
		{
			if (NODES[i].degree == 1)
			{
				NODES[i].next = list_first;
				list_first = i;
			}
		}

		while (list_first >= 0)
		{
			i = list_first;
			list_first = NODES[i].next;
			if (NODES[i].degree != 1) continue;

			for (dir=0; dir<2; dir++)
			{
				for (e=NODES[i].first[dir]; e>=0; e=EDGES[e].next[dir]) { if (map[e] >= 0) { j = edges[2*e+(1-dir)]; break; } }
				if (e >= 0) break;
			}
			if (e < 0) { printf("Fatal error in PM::reduce\n"); exit(1); }
			map[e] = -1; E --;
			NODES[i].degree --;
			NODES[j].degree --;
			for (dir=0; dir<2; dir++)
			{
				for (e=NODES[j].first[dir]; e>=0; e=EDGES[e].next[dir])
				{
					if (map[e] < 0) continue;
					i = edges[2*e+(1-dir)];
					map[e] = -2; E --;
					NODES[i].degree --;
					NODES[j].degree --;
					if (NODES[i].degree == 1)
					{
						NODES[i].next = list_first;
						list_first = i;
					}
				}
			}
			if (NODES[j].degree != 0) { printf("Fatal error in PM::reduce\n"); exit(1); }
		}
		for (i=0; i<node_num; i++)
		{
			if (NODES[i].degree > 0) NODES[i].degree = N ++;
			else NODES[i].degree = -1;
		}
		delete [] EDGES;

		///////

		PM* Q = new PM(N, E);
		E = 0;
		for (e=0; e<edge_num; e++)
		{
			if (map[e] >= 0)
			{
				map[e] = E ++;
				Q->AddEdge(NODES[edges[2*e]].degree, NODES[edges[2*e+1]].degree);
			}
		}

		delete [] NODES;

		Q->f = f->shrink(E);
		Function::SparseMap::Term T;
		T.c = 1;
		for (e=0; e<edge_num; e++)
		{
			if (map[e] >= -1)
			{
				T.v = map[e];
				f->shrink_add_term(Q->f, e, &T);
			}
		}

		Q->shrink_mapping = map;
		Q->f_method = f_method;
		return Q;
	}

#else

	Problem* reduce(int* map)
	{
		int i, j, k, e;
		int* node_map = new int[node_num + edge_num];
		int* edge_list = node_map + node_num;
		memset(node_map, 0, node_num*sizeof(int));
		int N = 0, E0 = 0, E = 0;
		for (e=0; e<edge_num; e++)
		{
			if (map[e] >= 0)
			{
				node_map[edges[2*e  ]] ++;
				node_map[edges[2*e+1]] ++;
				edge_list[E0 ++] = e;
			}
		}
		for (k=0; k<E0; k++)
		{
			e = edge_list[k]; i = edges[2*e]; j = edges[2*e+1];
			if (node_map[i] == 1 && node_map[j] == 1)
			{
				map[e] = -1;
				node_map[i] = node_map[j] = -1;
			}
			else
			{
				edge_list[E ++] = e;
			}
		}
		for (i=0; i<node_num; i++)
		{
			if (node_map[i] > 0)
			{
				node_map[i] = N ++;
			}
		}

		PM* P0 = new PM(N, E);

		for (k=0; k<E; k++)
		{
			e = edge_list[k]; i = edges[2*e]; j = edges[2*e+1];
			map[e] = k;
			P0->AddEdge(node_map[i], node_map[j]);
		}
		P0->f = f->shrink(P0->edge_num);
		Function::SparseMap::Term T;
		T.c = 1;
		for (e=0; e<edge_num; e++)
		{
			if (map[e] >= -1)
			{
				T.v = map[e];
				f->shrink_add_term(P0->f, e, &T);
			}
		}

		delete [] node_map;

		P0->shrink_mapping = map;
		P0->f_method = f_method;
		return P0;
	}
#endif

	Atom shrink_atom(Problem* _P, Atom _a)
	{
		PM* P = (PM*)_P;
		if (!P || P==this) return _a;
		int* map = P->shrink_mapping;
		int* a = (int*)_a;
		int bnum = P->node_num/2;
		int* b = new int[bnum];
		int i, blen = 0;
		for (i=0; i<node_num/2; i++)
		{
			int p = map[a[i]];
			if (p == -2) { delete [] b; return NULL; } //{ printf("error #1 in shrink_atom\n"); exit(1); }
			if (p >= 0)
			{
				if (blen >= bnum) { printf("error #2 in shrink_atom\n"); exit(1); }
				b[blen ++] = p;
			}
		}
		if (blen != bnum) { printf("error #3 in shrink_atom\n"); exit(1); }
		return b;
	}

	Atom unshrink_atom(Problem* _A, Atom _a)
	{
		PM* A = (PM*)_A;
		if (!A || A==this) return _a;
		int* map = A->shrink_mapping;
		int* a = (int*)_a;
		int* a_last = a + A->node_num/2;
		int bnum = node_num / 2, blen = 0;
		int* b = new int[bnum];
		int e;
		for (e=0; e<edge_num; e++)
		{
			if (map[e] == -2) continue;
			if (map[e] >= 0)
			{
				if (a >= a_last || map[e] != *a) continue;
				a ++;
			}
			if (blen >= bnum) { printf("error #1 in unshrink_atom\n"); exit(1); }
			b[blen ++] = e;
		}
		if (blen != bnum) { printf("error #2 in unshrink_atom\n"); exit(1); }
		delete [] ((int*)_a);

		return b;
	}

	Atom reshape_atom(Problem* A, Problem* B, Atom a) override
	{
		a = unshrink_atom(A, a);
		return shrink_atom(B, a);
	}

	void shrink_x(Problem* _A, ExtendedVector x)
	{
		PM* A = (PM*)_A;
		int* map = A->shrink_mapping;
		int e;
		for (e=0; e<edge_num; e++)
		{
			if (map[e] >= 0) x[map[e]] = x[e];
		}
	}

	void unshrink_x(Problem* _A, ExtendedVector x)
	{
		PM* A = (PM*)_A;
		double* xA = new double[A->edge_num];
		memcpy(xA, x, A->edge_num*sizeof(double));
		int* map = A->shrink_mapping;
		int e;
		for (e=0; e<edge_num; e++)
		{
			if (map[e] == -2) x[e] = 0;
			else if (map[e] == -1) x[e] = 1;
			else x[e] = xA[map[e]];
		}
		delete [] xA;
	}

	void reshape_x(Problem* A, Problem* B, ExtendedVector x) override
	{
		if (A == this) A = NULL;
		if (B == this) B = NULL;
		if (A) unshrink_x(A, x);
		if (B) shrink_x(B, x);

		if (B) ((PM*)B)->f->extend_x(x);
		else f->extend_x(x);
	}

	void PrintAtom(Atom _a) override
	{
	}

private:
	int f_method;
	PerfectMatching* P;
	double* g_prev;
	int* edges; // of size 2*edge_num
	int node_num;
	int edge_num, edge_num_max;
	Function* f;

	int* shrink_mapping; // of size edge_num
};


#endif
