#ifndef __SIMPLEX__
#define __SIMPLEX__

#include <string.h>
#include <stdlib.h>
#include "../problem.h"
#include "function.h"

class SIMPLEX : public Problem
{
public:
	SIMPLEX(int _n) : n(_n), shrink_mapping(NULL), f_method(0)
	{
	}
	~SIMPLEX()
	{
		if (shrink_mapping) delete [] shrink_mapping;
	}

	int get_n() override
	{
		return n;
	} 

	int get_N() override
	{
		return f->get_N();
	}

	void destroy_atom(Atom a) override
	{
		delete [] ((int*)a);
	}

	Atom get_init_atom()
	{
		int* a = new int[1];
		a[0] = rand() % n;
		return a;
	}

	Atom get_lowest_atom()
	{
		int* a = new int[1];
		a[0] = 0;
		return a;
	}

	void SetRandomFunction(bool uniform_A, int m, double sigma, double tau, long unsigned int seed, int _f_method=0)
	{
		f_method = _f_method;
		f = new Function(n, m, f_method);
		Atom _x0 = get_lowest_atom();
		double* x0 = new double[f->get_N()];
		atom_to_extended_vector(_x0, x0);
		destroy_atom(_x0);
		f->SetRandom(uniform_A, x0, sigma, seed);
		f->Scale(1 / tau);
		delete [] x0;
	}

	void SetRandomFunctionSkewed(bool uniform_A, int m, double sigma, long unsigned int seed, int _f_method=0)
	{
		f_method = _f_method;
		f = new Function(n, m, f_method);
		f->SetRandomSkewed(uniform_A, sigma, seed);
	}

	void atom_to_extended_vector(Atom _a, ExtendedVector x) override
	{
		int* a = (int*)_a;
		memset(x, 0, n*sizeof(double));
		x[a[0]] = 1;

		if (f_method > 0)
		{
			double* y = x + n;
			double** A = f->A;
			double* b = f->b;
			int m = f->m;
			// set y = Ax-b

			int i = a[0], d;
			for (d=0; d<m; d++)
			{
				double sum = -b[d];
				sum += A[d][i];
				y[d] = sum;
			}

			//int i, d;
			//for (d=0; d<m; d++)
			//{
			//	double sum = -b[d];
			//	for (i=0; i<n; i++) sum += A[d][i]*x[i];
			//	y[d] = sum;
			//}
		}
	}

	void add_atom_to_extended_vector(double gamma, Atom _a, ExtendedVector x) override
	{
		int* a = (int*)_a;
		x[a[0]] += gamma;

		if (f_method > 0)
		{
			double* y = x + n;
			double** A = f->A;
			double* b = f->b;
			int m = f->m;
			// set y += Ax-b

			int i = a[0], d;
			for (d=0; d<m; d++)
			{
				double sum = -b[d];
				sum += A[d][i];
				y[d] += gamma*sum;
			}
		}
	}

	double dot_product(Atom _a, Vector x) override
	{
		int* a = (int*)_a;
		return x[a[0]];
	}

	int compare_atoms(Atom _a, Atom _b) override
	{
		int* a = (int*) _a;
		int* b = (int*) _b;
		if (a[0] < b[0]) return -1;
		if (a[0] > b[0]) return +1;
		return 0;
	}

	double eval(ExtendedVector x) override
	{
		return f->eval(x);
	}

	void grad(ExtendedVector x, Vector g, double* fx) override
	{
		f->grad(x, g, fx);
	}

	double line_search(ExtendedVector x, ExtendedVector y, double* delta_f, double* delta_f1, double* delta_f_unconstrained) override
	{
		return f->line_search(x, y, delta_f, delta_f1, delta_f_unconstrained);
	}

	Atom min_oracle(Vector g, double time_limit) override
	{
		int i, i_opt = 0;
		double g_opt = g[0];
		for (i=1; i<n; i++)
		{
			if (g_opt > g[i]) { g_opt = g[i]; i_opt = i; }
		}
		int* a = new int[1];
		a[0] = i_opt;
		return a;
	}

	Atom minmax_oracle(Vector g, Vector x, Atom& v, double time_limit) override
	{
		int i, i_opt = -1;
		double g_opt = -1e300;
		for (i=0; i<n; i++)
		{
			if (x[i] > 0)
			{
				if (x[i] < 1e-10)
				{
					// renormalize
					x[i] = 0;
					double sum = 0;
					for (i=0; i<n; i++) sum += x[i];
					for (i=0; i<n; i++) x[i] /= sum;
					f->extend_x(x);
					return minmax_oracle(g, x, v, time_limit);
				}
				if (g_opt < g[i]) { g_opt = g[i]; i_opt = i; }
			}
		}
		int* a = new int[1];
		a[0] = i_opt;
		v = a;
		return min_oracle(g, time_limit);
	}

	double move_DICG(Vector x, Atom _s, Atom _v) override
	{
		int* s = (int*)_s;
		int* v = (int*)_v;
		double alpha = x[v[0]];
		x[v[0]] = 0;
		x[s[0]] += alpha;

		if (f_method > 0)
		{
			double* y = x + n;
			double** A = f->A;
			double* b = f->b;
			int m = f->m;

			int i = v[0], j=s[0], d;
			for (d=0; d<m; d++)
			{
				y[d] += alpha*(A[d][j] - A[d][i]);
			}
		}

		return alpha;
	}

	Problem* reduce(Vector g, ExtendedVector x, Atom _s, ReducePolicy& policy, double time_limit) override
	{
		int* s = (int*)_s;

		int i, nn;
		int* map = new int[n];

		for (i=nn=0; i<n; i++)
		{
			if (x[i] != 0 || i == s[0])
			{
				map[i] = nn ++;
			}
			else
			{
				map[i] = -1;
			}
		}

		if (!policy.is_minimal())
		{
			double* slacks = new double[n];
			int slack_num = 0;
			for (i=0; i<n; i++)
			{
				if (map[i] < 0) slacks[slack_num ++] = g[i];
			}
			assert(slack_num == n - nn);
			int m = policy.get_size(n, nn) - nn - 1;
			if (m >= 0)
			{
				std::nth_element(slacks, slacks+m, slacks+slack_num);
				double slack_threshold = slacks[m];
				nn = 0;
				for (i=0; i<n; i++)
				{
					if (map[i] >= 0 || g[i] < slack_threshold) map[i] = nn ++;
				}
			}

			delete [] slacks;
		}

		SIMPLEX* Q = new SIMPLEX(nn);
		Q->f = f->shrink(nn, map);
		Q->shrink_mapping = map;
		return Q;
	}

	Atom shrink_atom(Problem* _B, Atom _a)
	{
		SIMPLEX* B = (SIMPLEX*)_B;
		if (!B || B==this) return _a;
		int* a = (int*)_a;
		int* map = B->shrink_mapping;
		if (map[a[0]] < 0) { destroy_atom(_a); return NULL; }  //{ printf("error #1 in shrink_atom\n"); exit(1); }
		a[0] = map[a[0]];
		return _a;
	}
	Atom unshrink_atom(Problem* _A, Atom _a)
	{
		SIMPLEX* A = (SIMPLEX*)_A;
		if (!A || A==this) return _a;
		int* a = (int*)_a;
		int* map = A->shrink_mapping;
		int i;
		for (i=0; i<n; i++)
		{
			if (map[i] == a[0]) { a[0] = i; return _a; }
		}
		printf("unshrinking failed: shrinking should not be used with type-II algorithms\n"); exit(1);
	}
	Atom reshape_atom(Problem* A, Problem* B, Atom a) override
	{
		if (A == this) A = NULL;
		if (B == this) B = NULL;
		if (A) a = unshrink_atom(A, a);
		if (B) a = shrink_atom(B, a);
		return a;
	}
	void reshape_x(Problem* A, Problem* B, ExtendedVector x) override
	{
		if (A == this) A = NULL;
		if (B == this) B = NULL;
		if (A) unshrink_x(A, x);
		if (B) shrink_x(B, x);

		if (B) ((SIMPLEX*)B)->f->extend_x(x);
		else f->extend_x(x);
	}
	void shrink_x(Problem* _B, ExtendedVector x)
	{
		SIMPLEX* B = (SIMPLEX*)_B;
		if (!B || B==this) return;
		int* map = B->shrink_mapping;
		int i;
		for (i=0; i<n; i++)
		{
			if (map[i] >= 0) x[map[i]] = x[i];
		}
	}
	void unshrink_x(Problem* _A, ExtendedVector x)
	{
		SIMPLEX* A = (SIMPLEX*)_A;
		if (!A || A==this) return;
		int* map = A->shrink_mapping;
		double* y = new double[A->n];
		memcpy(y, x, A->n*sizeof(double));
		int i;
		for (i=0; i<n; i++)
		{
			x[i] = (map[i] >= 0) ? y[map[i]] : 0;
		}
	}

	void test_x_consistency(double* x) override
	{
		f->test_x_consistency(x);
	}

private:
	int n; 
	Function* f;
	int f_method;
	int* shrink_mapping; // of size n

};


#endif
