#ifndef __FUNCTION_H__
#define __FUNCTION_H__

#include <stdio.h>
#include <stdlib.h>
#include <random>
#include "../utils.h"

// function f(x) = || A x - b ||^2

class Function
{
public:
	Function(int _n, int _m, int _f_method=0) : n(_n), m(_m), f_method(_f_method)
	{
		int d;
		A = new double*[m];
		A[0] = new double[m*n];
		for (d=1; d<m; d++) A[d] = A[d-1] + n;
		b = new double[m];
		buf_m = new double[m];
		buf_m2 = new double[m];
		buf_n = new double[n];
		memset(A[0], 0, m*n*sizeof(double));
		memset(b, 0, m*sizeof(double));
		const_term = 0;
	}
	~Function()
	{
		delete [] A[0];
		delete [] A;
		delete [] b;
		delete [] buf_m;
		delete [] buf_m2;
		delete [] buf_n;
	}
	int get_N()
	{
		if (f_method == 0) return n;
		return n + m;
	}

	void Scale(double c)
	{
		int d, i;
		for (d=0; d<m; d++)
		{
			for (i=0; i<n; i++) A[d][i] *= c;
		}
	}

	void Print()
	{
		int d, i;
		for (d=0; d<m; d++)
		{
			if (d == 0) printf("A=[");
			else        printf("   ");
			for (i=0; i<n; i++)
			{
				printf("%f", A[d][i]);
				if (i==n-1 && d==m-1) printf("]\n");
				else if (i==n-1) printf(";\n");
				else printf(", ");
			}
		}
		printf("b=[");
		for (d=0; d<m; d++)
		{
			printf("%f", b[d]);
			if (d==m-1) printf("]\n");
			else printf(", ");
		}
		if (const_term != 0) printf("const_term=%f\n", const_term);
	}

	void SetRandom(bool uniform_A, double* x0, double sigma, long unsigned int seed)
	{
		int d, i;

		std::default_random_engine gen(seed);
		std::normal_distribution<double> NORMAL(0, 1);
		std::uniform_real_distribution<> UNIFORM(0.0, 1.0);

		for (d=0; d<m; d++)
		for (i=0; i<n; i++)
		{
			if (uniform_A) A[d][i] = UNIFORM(gen); 
			else           A[d][i] = NORMAL(gen);
		}

		Ax(x0, b);

		if (sigma > 0)
		{
			std::normal_distribution<double> NORMAL(0, sigma);
			for (d=0; d<m; d++)
			{
				b[d] += NORMAL(gen);
			}
		}
	}

	void SetRandomSkewed(bool uniform_A, double sigma, long unsigned int seed)
	{
		if (n&1) { printf("Error in SetRandomSkewed: N is odd\n"); exit(1); }

		int d, i;

		double* x0 = new double[n];

		std::default_random_engine gen(seed);
		std::uniform_real_distribution<> UNIFORM(0.0, 1.0);
		std::normal_distribution<double> NORMAL(0, 1);

		double tau = 0;
		for (i=0; i<n/2; i++)
		{
			//x0[i] = NORMAL(gen);
			x0[i] = (i<n/50) ? 1 : 0;
			tau += fabs(x0[i]);
		}

		for (d=0; d<m; d++)
		for (i=0; i<n/2; i++)
		{
			if (uniform_A) A[d][i] = UNIFORM(gen); 
			else           A[d][i] = NORMAL(gen);
			A[d][i+n/2] = -A[d][i]; 
		}

		//////////////////////

		n /= 2; Ax(x0, b); n *= 2;

		delete [] x0;

		if (sigma > 0)
		{
			for (d=0; d<m; d++)
			{
				b[d] += sigma*NORMAL(gen);
			}
		}

		Scale(tau);
	}

	void extend_x(double* x)
	{
		if (f_method > 0)
		{
			Ax_b(x, x+n);
		}
	}
	void test_x_consistency(double* x)
	{
		if (f_method > 0)
		{
			double* y = x+n;
			double* c = buf_m;
			Ax_b(x, c);
			int d;
			for (d=0; d<m; d++)
			{
				if (fabs(y[d] - c[d]) > 1e-10)
				{
					printf("Inconsistent x\n");
					exit(1);
				}
			}
		}
	}

	double eval(double* x)
	{
		double* c = buf_m;

		if (f_method == 0) Ax_b(x, c); // set c = Ax-b
		else c = x + n;

		double sum = norm_squared(c, m) + const_term;
		return sum;
	}

	void grad(double* x, double* g, double* f)
	{
		int i, d;

		double* c = buf_m;

		if (f_method == 0) Ax_b(x, c); // set c = Ax-b
		else c = x + n;

		if (f)
		{
			*f = norm_squared(c, m) + const_term;
		}
		memset(g, 0, n*sizeof(double));
		for (d=0; d<m; d++)
		{
			for (i=0; i<n; i++) g[i] += 2*A[d][i]*c[d];
		}
	}

	void grad_numerical(double* x, double* g, double eps=1e-6)
	{
		int i;
		double F0 = eval(x);
		for (i=0; i<n; i++)
		{
			double x_old = x[i];
			x[i] += eps;
			g[i] = (eval(x) - F0) / eps;
			x[i] = x_old;
		}
	}

	double line_search(double* x, double* y, double* delta_f, double* delta_f1, double* delta_f_unconstrained, bool clip=true)
	{
		int i;
		double* c = buf_m;
		double* d = buf_m2;
		double* yx = buf_n;

		if (f_method == 0) Ax_b(x, c); // set c = Ax-b
		else c = x + n;

		for (i=0; i<n; i++) yx[i] = y[i] - x[i];
		Ax(yx, d); // set d = A(y-x)

		double AA = norm_squared(d, m);
		double BB = dot_product(c, d, m);

		double gamma = (AA > 0) ? -BB/AA : 1;
		if (delta_f_unconstrained)  *delta_f_unconstrained  = (AA*gamma+2*BB)*gamma;
		if (clip)
		{
			if (gamma < 0) gamma = 0;
			if (gamma > 1) gamma = 1;
		}

		if (delta_f)  *delta_f  = (AA*gamma+2*BB)*gamma;
		if (delta_f1) *delta_f1 = AA+2*BB;

		return gamma;
	}
		
	//////////////////////////////////////////////////////
	void test_grad(double* x)
	{
		int i;
		double* g = new double[n];

		printf("x              = ");
		for (i=0; i<n; i++) printf("%f ", x[i]);
		printf("\n");

		printf("grad           = ");
		grad(x, g, NULL);
		for (i=0; i<n; i++) printf("%f ", g[i]);
		printf("\n");

		printf("grad_numerical = ");
		grad_numerical(x, g);
		for (i=0; i<n; i++) printf("%f ", g[i]);
		printf("\n");

		delete [] g;
	}

	void test_line_search(double* x, double* y)
	{
		int i;
		double* z = new double[2*n];
		double* g = z + n;

		double delta_f, delta_f1, delta_f_unconstrained;
		double gamma = line_search(x, y, &delta_f, &delta_f1, &delta_f_unconstrained, false);
		for (i=0; i<n; i++) z[i] = (1-gamma)*x[i] + gamma*y[i];
		grad(z, g, NULL);
		double p = 0;
		for (i=0; i<n; i++) p += (y[i] - x[i])*g[i];
		printf("line_search: gamma=%f, <gz,y-x>=%f\n", gamma, p);
		printf("f(z)=[%f, %f], f(y)=[%f, %f]\n", eval(z), eval(x)+delta_f, eval(y), eval(x)+delta_f1);
		delete [] z;
	}

	struct SparseMap
	{
		struct Term { int v; double c; };
		Term* terms;
		int num;
	};
	Function* shrink(int nn, SparseMap* mapping=NULL) // mapping is of size n, the new problem is of size nn
	{
		int d, i;
		Function* ff = new Function(nn, m, f_method);
		for (d=0; d<m; d++) ff->b[d] = b[d];

		if (mapping)
		{
			SparseMap::Term* t;

			for (d=0; d<m; d++)
			{
				for (i=0; i<n; i++)
				{
					for (t=mapping[i].terms; t<mapping[i].terms+mapping[i].num; t++)
					{
						if (t->v >= ff->n) { printf("invalid mapping in Function::shrink\n"); exit(1); }
						double c = A[d][i]*t->c;
						if (t->v<0) ff->b[d] -= c;
						else ff->A[d][t->v] += c;
					}
				}
			}
		}

		return ff;
	}
	void shrink_add_term(Function* ff, int i, SparseMap::Term* t) // alternatively, one can set mapping=NULL in shrink() and then add terms one-by-one
	{
		int d;
		for (d=0; d<m; d++)
		{
			if (t->v >= ff->n) { printf("invalid mapping in Function::shrink\n"); exit(1); }
			double c = A[d][i]*t->c;
			if (t->v<0) ff->b[d] -= c;
			else ff->A[d][t->v] += c;
		}
	}

	Function* shrink(int nn, int* map) // new problem is of size nn, map is of size n. If map[i] = ii >= 0 then old variable i is equal to variable ii in the contracted problem
	{
		int d, i;
		Function* ff = new Function(nn, m);
		for (d=0; d<m; d++) ff->b[d] = b[d];

		for (d=0; d<m; d++)
		{
			for (i=0; i<n; i++)
			{
				if (map[i] >= 0) ff->A[d][map[i]] += A[d][i];
			}
		}
		return ff;
	}
										
	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////


//private:
	int f_method;
	int n, m;
	double** A; // A[d][i] where d\in[Adim], i\in[n]
	double* b;
	double const_term;
	double* buf_m;
	double* buf_m2;
	double* buf_n;

	void Ax(double* x, double* y) // set y=Ax
	{
		int i, d;
		for (d=0; d<m; d++)
		{
			double sum = 0;
			for (i=0; i<n; i++) sum += A[d][i]*x[i];
			y[d] = sum;
		}
	}
	void Ax_b(double* x, double* y) // set y=Ax-b
	{
		int i, d;
		for (d=0; d<m; d++)
		{
			double sum = -b[d];
			for (i=0; i<n; i++) sum += A[d][i]*x[i];
			y[d] = sum;
		}
	}
};

/*
void TestFunction()
{
	int i;
	const int n = 5, m = 3;
	bool is_reduced = true;
	Function F(n, m, is_reduced);
	double x[n], y[n];
	for (i=0; i<n; i++) x[i] = i%2;
	F.SetRandom(x, 0.1);
	for (i=0; i<n; i++)
	{
		x[i] = (i+1)%2;
		y[i] = 2*i;
	}
	F.test_grad(x);
	F.test_line_search(x, y);
}
*/

#endif
