#ifndef __FUNCTIONQ_H__
#define __FUNCTIONQ_H__

#include <stdio.h>
#include <stdlib.h>
#include <random>
#include "../utils.h"

// f(x) = 0.5 x^T Q x + b^T x, for a symmetric SDP matrix Q

class FunctionQ
{
public:
	FunctionQ(int _n) : n(_n)
	{
		int d;
		b = new double[n];
		Q = new double*[n];
		Q[0] = new double[n*n];
		for (d=1; d<n; d++) Q[d] = Q[d-1] + n;
		buf_n = new double[n];
		buf2_n = new double[n];
		memset(b, 0, n*sizeof(double));
		memset(Q[0], 0, n*n*sizeof(double));
		const_term = 0;
	}
	~FunctionQ()
	{
		delete [] b;
		delete [] Q[0];
		delete [] Q;
		delete [] buf_n;
		delete [] buf2_n;
	}

	void SetRandom(long unsigned int seed)
	{
		std::default_random_engine gen(seed);
		std::uniform_real_distribution<> UNIFORM(0.0, 1.0);

		double* c = buf_n;
		int i, j, r;

		for (r=0; r<n/2; r++)
		{
			for (i=0; i<n; i++) c[i] = UNIFORM(gen); 
			for (i=0; i<n; i++) 		
			for (j=0; j<n; j++)
				Q[i][j] += c[i]*c[j];
		}

		for (i=0; i<n; i++) b[i] = UNIFORM(gen);

		/*
		for (r=0; r<1; r++)
		{
			for (i=0; i<n; i++) c[i] = (int)(10*UNIFORM(gen)); 
			for (i=0; i<n; i++) 		
			for (j=0; j<n; j++)
				Q[i][j] += c[i]*c[j];
		}
		*/

		for (i=0; i<n; i++) b[i] = (int)(10*UNIFORM(gen));
	}

	void Print()
	{
		int d, i;
		for (d=0; d<n; d++)
		{
			if (d == 0) printf("Q=[");
			else        printf("   ");
			for (i=0; i<n; i++)
			{
				printf("%f", Q[d][i]);
				if (i==n-1 && d==n-1) printf("]\n");
				else if (i==n-1) printf(";\n");
				else printf(", ");
			}
		}
		for (d=0; d<n; d++)
		{
			if (d == 0) printf("b=[");
			else        printf("   ");
			printf("%f, ", b[d]);
		}
		printf("]\n");
		if (const_term != 0) printf("const_term=%f\n", const_term);
	}

	double eval(double* x)
	{
		int i;
		double* c = buf_n;
		Qx(x, c); // set c = Qx
		double sum = const_term;
		for (i=0; i<n; i++) sum += (0.5*c[i]+b[i])*x[i];
		return sum;
	}

	void grad(double* x, double* g, double* f, Function* ff=NULL)
	{
		int i;
		Qx(x, g); // set g = Qx
		if (f)
		{
			double sum = const_term;
			for (i=0; i<n; i++) sum += (0.5*g[i]+b[i])*x[i];
			*f = sum;
		}
		for (i=0; i<n; i++) g[i] += b[i];
	}

	void grad_numerical(double* x, double* g, double eps=1e-6)
	{
		int i;
		double F0 = eval(x);
		for (i=0; i<n; i++)
		{
			double x_old = x[i];
			x[i] += eps;
			g[i] = (eval(x) - F0) / eps;
			x[i] = x_old;
		}
	}

	double line_search(double* x, double* y, double* delta_f, double* delta_f1, double* delta_f_unconstrained, bool clip=true)
	{
		int i;
		double* c = buf_n;
		double* d = buf2_n;
		for (i=0; i<n; i++) d[i] = y[i] - x[i];
		Qx(d, c); // set c = Q(y-x)
		double AA = 0, BB = 0;
		for (i=0; i<n; i++)
		{
			AA += d[i]*c[i];
			BB += x[i]*c[i] + b[i]*d[i];
		}
		AA *= 0.5;
		BB *= 0.5;

		double gamma = (AA > 0) ? -BB/AA : 1;
		if (delta_f_unconstrained)  *delta_f_unconstrained  = (AA*gamma+2*BB)*gamma;
		if (clip)
		{
			if (gamma < 0) gamma = 0;
			if (gamma > 1) gamma = 1;
		}

		if (delta_f)  *delta_f  = (AA*gamma+2*BB)*gamma;
		if (delta_f1) *delta_f1 = AA+2*BB;

		return gamma;
	}
		
	//////////////////////////////////////////////////////
	void test_grad(double* x)
	{
		int i;
		double* g = new double[n];

		printf("x              = ");
		for (i=0; i<n; i++) printf("%f ", x[i]);
		printf("\n");

		printf("grad           = ");
		grad(x, g, NULL);
		for (i=0; i<n; i++) printf("%f ", g[i]);
		printf("\n");

		printf("grad_numerical = ");
		grad_numerical(x, g);
		for (i=0; i<n; i++) printf("%f ", g[i]);
		printf("\n");

		delete [] g;
	}

	void test_line_search(double* x, double* y)
	{
		int i;
		double* z = new double[2*n];
		double* g = z + n;

		double delta_f, delta_f1, delta_f_unconstrained;
		double gamma = line_search(x, y, &delta_f, &delta_f1, &delta_f_unconstrained, false);
		for (i=0; i<n; i++) z[i] = (1-gamma)*x[i] + gamma*y[i];
		grad(z, g, NULL);
		double p = 0;
		for (i=0; i<n; i++) p += (y[i] - x[i])*g[i];
		printf("line_search: gamma=%f, <gz,y-x>=%f\n", gamma, p);
		printf("f(z)=[%f, %f], f(y)=[%f, %f]\n", eval(z), eval(x)+delta_f, eval(y), eval(x)+delta_f1);
		delete [] z;
	}

	FunctionQ* shrink(int nn, int* map) // new problem is of size nn, map is of size n. If map[i] = ii >= 0 then old variable i is equal to variable ii in the contracted problem
	{
		int i, j;
		FunctionQ* ff = new FunctionQ(nn);
		for (i=0; i<n; i++)
		{
			int ii = map[i];
			if (ii < 0) continue;
			for (j=0; j<n; j++)
			{
				int jj = map[j];
				if (jj < 0) continue;
				ff->Q[ii][jj] += Q[i][j];
			}
			ff->b[ii] += b[i];
		}
		return ff;
	}

	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////
	//////////////////////////////////////////////////////

//private:
	int n;
	double** Q; // Q[i][j]
	double* b; // Q[i][j]
	double const_term;
	double* buf_n;
	double* buf2_n;

	void Qx(double* x, double* y) // set y=Ax
	{
		int i, d;
		for (d=0; d<n; d++)
		{
			double sum = 0;
			for (i=0; i<n; i++) sum += Q[d][i]*x[i];
			y[d] = sum;
		}
	}
};

/*
void TestFunction()
{
	int i;
	const int n = 5;
	FunctionQ F(n);
	double x[n], y[n];
	F.SetRandom(1);
	for (i=0; i<n; i++)
	{
		x[i] = (i+1)%2;
		y[i] = 2*i;
	}
	F.test_grad(x);
	F.test_line_search(x, y);
	exit(1);
}
*/

#endif
