/*
    Copyright Vladimir Kolmogorov vnk@ist.ac.at 2020-2021

    This file is part of RECURSIVE-FW.

    POTTS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "problem.h"
#include "solver.h"
#include "utils.h"


////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

void Solver::BCG()
{
	double Phi = FW_INFTY;

	P->init_grad(g);

	bool simplex_allowed = true;
	while ( 1 )
	{
		BCGstep(Phi, simplex_allowed, NO_TIME_LIMIT);
		if (!L->keep_going()) break;
	}
}

void Solver::BCGstep(double& Phi, bool& simplex_allowed, double time_limit)
{
	const char* type = "";
	bool is_drop = false, is_add = false;
	int atom_num_old = atom_num;
	double f, gap=FW_INFTY, W=FW_INFTY, gamma;
	int p_min, p_max;
	double c_min, c_max;

	P->grad(x, g, &f); L->log_f(f);

	L->log_Phi(Phi);

	compute_c_products(P, p_min, c_min, p_max, c_max); L->log_wminus(c_max-c_min);
	if (c_max - c_min >= Phi / (2*options.K1) && simplex_allowed)
	{
		double improvement;
		if (!simplex_step(improvement))
		{
			simplex_allowed = false;
			if (options.terminate_on_simplex_failure) L->set_terminate();
		}
		if (atom_num < atom_num_old) is_drop = true;
		if (simplex_allowed) type = (is_drop) ? "simplex-" : "simplex";
		else                 type = (is_drop) ? "simplex- (?)" : "simplex (?)";
	}
	else if (dot_product(x, g, n) - c_min >= Phi / (2*options.K2) && simplex_allowed)
	{
		Problem::Atom s = atoms[p_min].atom;
		P->atom_to_extended_vector(s, xs);
		gamma = P->line_search(x, xs, NULL, NULL, NULL); L->log_line_search(gamma);
		interpolate(x, xs, gamma, x, N);
		update_weights_FW_step(gamma, s);
		type = "cache";
	}
	else
	{
		Problem::Atom s = P->min_oracle(g, time_limit); L->log_min_oracle();
		if (!s)
		{
			L->log_str("oracle-failed");
			L->set_terminate_current_level();
			return;
		}
		P->atom_to_extended_vector(s, xs);
		gap = dot_product_diff(x, xs, g, n); L->log_gap(gap);
		W = c_max - P->dot_product(s, g); L->log_w(W);
		bool half_step = false, accept_step = true;
		if (gap < Phi/2)
		{
			Phi /= 2;
			if (Phi > W) Phi = W;
			if (!options.BCG_accept_half_step) accept_step = false;
			half_step = true;
		}
		if (accept_step)
		{
			gamma = P->line_search(x, xs, NULL, NULL, NULL); L->log_line_search(gamma);
			interpolate(x, xs, gamma, x, N);
			AtomIndex ps = find_and_replace_atom(s);
			if (ps < 0)
			{
				simplex_allowed = true;
				is_add = true;
			}
			update_weights_FW_step(gamma, s);
		}
		if (half_step) type = (is_add) ? "oracle-half+" : "oracle-half";
		else           type = (is_add) ? "oracle+" : "oracle";
	}
	L->log_atom_num(atom_num);
	L->log_str(type);
}

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

void Solver::update_weights_FW_step(double gamma, Problem::Atom s)
{
	MyAtom* a;
	AtomIndex p, ps = -1;
	for (p=0; p<atom_num; p++) { if (atoms[p].atom == s) { ps = p; break; } }

	if (gamma <= 0) return;
	if (gamma >= 1)
	{
		for (p=0; p<atom_num; p++)
		{
			if (p != ps) P->destroy_atom(atoms[p].atom);
		}
		a = &atoms[0];
		a->alpha = 1;
		a->atom = s;
		atom_num = 1;
	}
	else
	{
		for (p=0; p<atom_num; p++) atoms[p].alpha *= 1-gamma;
		if (ps >= 0)
		{
			atoms[ps].alpha += gamma;
		}
		else
		{
			ps = add_atom(s);
			atoms[ps].alpha = gamma;
		}
	}
}

void Solver::compute_c_products(Problem* Q, int& p_min, double& c_min, int& p_max, double& c_max)
{
	int p;
	p_min = p_max = -1;
	c_min = c_max = 0;
	for (p=0; p<atom_num; p++)
	{
		MyAtom* a = &atoms[p];
		a->c = Q->dot_product(a->atom, g);
		if (p == 0 || c_min>a->c) { p_min = p; c_min = a->c; }
		if (p == 0 || c_max<a->c) { p_max = p; c_max = a->c; }
	}
}

bool Solver::simplex_step(double& improvement)
{
	AtomIndex p, q;
	MyAtom* a;

	double c_sum = 0;
	for (p=0; p<atom_num; p++) c_sum += atoms[p].c;
	c_sum /= atom_num;
	for (p=0; p<atom_num; p++) atoms[p].c -= c_sum;

	q = -1;
	double eta = 0;
	for (p=0; p<atom_num; p++)
	{
		a = &atoms[p];
		if (a->c > 0 && (q<0 || eta > a->alpha / a->c))
		{
			q = p;
			eta = a->alpha / a->c;
		}
	}

	if (q < 0)
	{
		return false;
	}

	c_sum = 0;
	for (p=0; p<atom_num; p++)
	{
		if (p == q) atoms[p].c = 0;
		else
		{
			atoms[p].c = atoms[p].alpha - eta*atoms[p].c;
			if (atoms[p].c < 0) atoms[p].c = 0;
			c_sum += atoms[p].c;
		}
	}
	if (c_sum <= 0)
	{
		return false;
	}
	for (p=0; p<atom_num; p++) atoms[p].c /= c_sum;

	memset(xs, 0, N*sizeof(double));
	for (p=0; p<atom_num; p++)
	{
		a = &atoms[p];
		P->add_atom_to_extended_vector(a->c, a->atom, xs);
	}

	double delta_f, delta_f1;
	double gamma = P->line_search(x, xs, &delta_f, &delta_f1, NULL);

	improvement = 0;

	if (delta_f1 <= 0)
	{
		gamma = 1;
		improvement = -delta_f1;
	}
	else
	{
		improvement = -delta_f;
		if (improvement <= 0) return false;
	}

	L->log_line_search(gamma);

	if (gamma < 1) q = -1;
	double alpha_sum = 0;
	for (p=atom_num-1; p>=0; p--)
	{
		a = &atoms[p];
		a->alpha = (1-gamma)*a->alpha + gamma*a->c;
		if (a->alpha <= 0 || p == q) remove_atom(p);
		else alpha_sum += a->alpha;
	}

	if (alpha_sum<1-1e-6 || alpha_sum>1+1e-6) { printf("error in simplex_step: alpha_sum = %f\n", alpha_sum); exit(1); }

	memset(x, 0, N*sizeof(double));
	for (p=0; p<atom_num; p++)
	{
		a = &atoms[p];
		a->alpha /= alpha_sum;
		P->add_atom_to_extended_vector(a->alpha, a->atom, x);
	}


	return true;
}



















void Solver::BCG_R()
{
	L->start_recursive(n);
	P->init_grad(g);
	BCG_R(0, 0, 1e300, _get_time() + 1e300);
}




void Solver::BCG_R(int depth, double T_parent, double Phi_parent, double T_limit)
{
	double W;
	double Phi = Phi_parent;
	Problem::Atom s;
	MyAtom* a;
	int t;

	double TIME_LIMIT;
	if (depth>0 && options.subproblem_time_limit>0) TIME_LIMIT = T_parent*options.subproblem_time_limit;
	else                                            TIME_LIMIT = NO_TIME_LIMIT;

	for (t=0; options.reduce_policy->reduce_allowed(depth); t++)
	{
		Problem* P0 = P;

		///////////////////////
		double T_start = _get_time();
		ComputeGapW_II(&W, &s);
		//if (Phi > W) Phi = W;
		Phi = W;
		double T_current = _get_time();
		double T = T_current - T_start;
		if ( !L->keep_going() || T_current > T_limit ) { P->destroy_atom(s); return; }
		double T_limit_R = T_limit;
		if (options.subproblem_time_limit > 0)
		{
			double TT = T_current + options.subproblem_time_limit*T;
			if (T_limit_R > TT) T_limit_R = TT;
		}
		///////////////////////

		if (depth > 0 && t > 0)
		{
			if ( Phi*Phi*T_parent <= Phi_parent*Phi_parent*T && t > 0) { P->destroy_atom(s); return; }
		}

		///////////////////////////////////////////////////////

		P = P0->reduce(g, x, s, *options.reduce_policy, TIME_LIMIT); 
		P0->destroy_atom(s);
		if (!P)
		{
			P = P0;
			return;
		}
		memset(x, 0, P->get_N()*sizeof(double));
		for (a=atoms; a<atoms+atom_num; a++)
		{
			a->atom = P0->reshape_atom(P0, P, a->atom);
			P->add_atom_to_extended_vector(a->alpha, a->atom, x);
		}

		n = P->get_n(); N = P->get_N(); L->log_shrink(n);
		P->init_grad(g);

		if (depth>0)
		{
			BCG_R(depth+1, T_parent, Phi_parent, T_limit_R);
		}
		else
		{
			BCG_R(depth+1, T, Phi, T_limit_R);
		}

		L->log_unshrink();
		n = P0->get_n(); N = P0->get_N(); 

		memset(x, 0, N*sizeof(double));
		for (a=atoms; a<atoms+atom_num; a++)
		{
			a->atom = P0->reshape_atom(P, NULL, a->atom);
			P0->add_atom_to_extended_vector(a->alpha, a->atom, x);
		}
		delete P;
		P = P0;
		P->init_grad(g);
	}

	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////

	int iter;
	bool simplex_allowed = true;
	for (iter=0; ; iter++)
	{
		double T_start = _get_time();
		//		double time_limit = options.subproblem_time_limit * Phi*Phi * T_root / (Phi_root*Phi_root);
		//		double W;
		BCGstep(Phi, simplex_allowed, NO_TIME_LIMIT);
		if (!L->keep_going()) break;
		double T = _get_time();
		if (T>T_limit) break;
		if (depth > 0 && Phi*Phi*T_parent < options.reduce_time_ratio*Phi_parent*Phi_parent*(T-T_start)) break;
	}
}


double Solver::ComputeGapW_II(double* W, Problem::Atom* s_ptr, double TIME_LIMIT)
{
	double f, gap, gx, gs;
	P->grad(x, g, &f); L->log_f(f);

	Problem::Atom s = P->min_oracle(g, TIME_LIMIT); L->log_min_oracle();
	if (!s) { L->log_str("oracle-failed"); return 0; }
	gs = P->dot_product(s, g);
	gx = dot_product(x, g, n);
	gap = gx - gs; L->log_gap(gap);
	if (s_ptr) *s_ptr = s;
	else P->destroy_atom(s);

	if (W)
	{
		int p_min, p_max;
		double c_min, c_max;
		compute_c_products(P, p_min, c_min, p_max, c_max); L->log_wminus(c_max-c_min);
		*W = c_max - gs; L->log_w(*W);
	}
	return gap;
}

