/*
    Copyright Vladimir Kolmogorov vnk@ist.ac.at 2020-2021

    This file is part of RECURSIVE-FW.

    POTTS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <algorithm>
#include "problem.h"
#include "solver.h"
#include "utils.h"

void Solver::CacheDICG()
{
	P->init_grad(g);

	while ( 1 )
	{
		double W=1e300;
		CacheDICGstep(W, NO_TIME_LIMIT);
		if (!L->keep_going()) break;
	}
}

void Solver::CacheFW()
{
	P->init_grad(g);

	while ( 1 )
	{
		double W=1e300;
		CacheFWstep(NO_TIME_LIMIT);
		if (!L->keep_going()) break;
	}
}

void Solver::CacheFWstep(double time_limit)
{
	double f, delta_f, delta_f1, delta_f_unconstrained, gamma;
	double gap;

	double T0 = _get_time();

	P->grad(x, g, &f); L->log_f(f);

	Problem::Atom s = P->min_oracle(g, NO_TIME_LIMIT); L->log_min_oracle();
	P->atom_to_extended_vector(s, xs);
	add_atom_to_cache(s);
	gap = dot_product_diff(x, xs, g, n); L->log_gap(gap);

	gamma = P->line_search(x, xs, &delta_f, &delta_f1, &delta_f_unconstrained); L->log_line_search(gamma);
	interpolate(x, xs, gamma, x, N);

	L->log_atom_num(atom_num);

	T0 = _get_time() - T0;
	double improvement0 = -delta_f_unconstrained;

	int cache_iter;
	for (cache_iter=0; cache_iter<options.cache_subiter_max; cache_iter++)
	{
		double improvement;
		double T = _get_time();
		if (!cache_step(improvement)) break;
		T = _get_time() - T;
		if (improvement * T0 <= options.cache_time_ratio*improvement0 * T) break;
		if (!L->keep_going()) break;
	}
	timestamp ++;
}

double Solver::CacheDICGstep(double& Phi, double time_limit)
{
	double f, delta_f, delta_f1, delta_f_unconstrained, gamma, alpha;

	double T0 = _get_time();
	P->grad(x, g, &f); L->log_f(f);
	Problem::Atom v;

	Problem::Atom s = P->minmax_oracle(g, x, v, time_limit); L->log_min_max_oracle();
	double gx = dot_product(x, g, n);
	double gs = P->dot_product(s, g);
	double gv = P->dot_product(v, g);

	double gap = gx - gs; L->log_gap(gap);
	double W = gv - gs; L->log_w(W);
	if (Phi > W) Phi = W;

	if (options.DICG_method == Options::PFW)
	{
		memcpy(xs, x, N*sizeof(double));
		alpha = P->move_DICG(xs, s, v); L->log_dicg_alpha(alpha);
	}
	else
	{
		if (gap >= gv - gx)
		{
			P->atom_to_extended_vector(s, xs);
			alpha = options.DICG_eps + 1.0;
		}
		else
		{
			P->atom_to_extended_vector(v, xs);
			alpha = P->move_away(x, xs); L->log_dicg_alpha(alpha);
		}
	}

	add_atom_to_cache(s);
	P->destroy_atom(v);

	gamma = P->line_search(x, xs, &delta_f, &delta_f1, &delta_f_unconstrained); L->log_line_search(gamma);
	interpolate(x, xs, gamma, x, N);

	if (alpha <= options.DICG_eps)
	{
		P->atom_to_extended_vector(s, xs);
		gamma = P->line_search(x, xs, NULL, NULL, NULL); L->log_line_search(gamma);
	}

	L->log_atom_num(atom_num);

	T0 = _get_time() - T0;
	double improvement0 = -delta_f_unconstrained;

	int cache_iter;
	for (cache_iter=0; cache_iter<options.cache_subiter_max; cache_iter++)
	{
		double improvement;
		double T = _get_time();
		if (!cache_step(improvement)) break;
		T = _get_time() - T;
		if (improvement * T0 <= options.cache_time_ratio*improvement0 * T) break;
		if (!L->keep_going()) break;
	}
	timestamp ++;

	return T0;
}


// find delta such that
// (c[0]-delta) + \sum_{i=1}^{n-1} \min { c[i]-delta, 0 } = 0

bool project_grad_to_simplex(int n, double* c) // at (1,0,...,0)
{
	int k;
	double* cc = c+n;
	memcpy(cc, c+1, (n-1)*sizeof(double));
	std::sort(cc, cc+n-1);
	double sum = c[0], delta = c[0];
	for (k=1; k<n; k++)
	{
		delta = cc[k-1];
		if (sum - k*delta <= 0) break;
		sum += cc[k-1];
	}
	if (k == 1) return false;
	delta = sum / k;

	c[0] -= delta;
	if (c[0] <= 0) return false;
	for (k=1; k<n; k++)
	{
		c[k] -= delta;
		if (c[k] >= 0) c[k] = 0;
		c[k] /= c[0];
	}
	c[0] = 1;

	sum = 0;
	for (k=1; k<n; k++) sum -= c[k];
	if (sum <= 0) return false;
	for (k=1; k<n; k++) c[k] /= sum;

	/*
	/// check
	sum = 0;
	for (k=0; k<n; k++) sum += c[k];
	if (fabs(sum) > 1e-10)
	{
		printf("Error: sum=%f\n", sum);
		exit(1);
	}
	*/

	return true;
}

bool Solver::cache_step(double& improvement)
{
	AtomIndex p;
	MyAtom* a;
	double f;
	int i;

	P->grad(x, g, &f); L->log_f(f);

	int p_min, p_max;
	double c_min, c_max;
	compute_c_products(P, p_min, c_min, p_max, c_max);
	double* c = new double[2*(atom_num+1)];
	c[0] = dot_product(x, g, n);
	for (p=0; p<atom_num; p++) c[p+1] = atoms[p].c;

	if (!project_grad_to_simplex(atom_num+1, c))
	{
		delete [] c;
		return false;
	}

	for (i=0; i<N; i++) xs[i] = (1-c[0])*x[i];
	for (p=0; p<atom_num; p++)
	{
		if (c[p+1] >= 0) continue;
		a = &atoms[p];
		P->add_atom_to_extended_vector(-c[p+1], a->atom, xs);
		atoms[p].timestamp = timestamp;
	}

	double delta_f;
	double gamma = P->line_search(x, xs, &delta_f, NULL, NULL); L->log_line_search(gamma);
	interpolate(x, xs, gamma, x, N);
	improvement = -delta_f;

	delete [] c;
	return true;
}
















void Solver::CacheDICG_R()
{
	L->start_recursive(n);
	P->init_grad(g);
	CacheDICG_R(0, 0, 1e300, _get_time() + 1e300);
}


double Solver::ComputeDICGimprovement(Problem::Atom* _s)
{
	double f;
	P->grad(x, g, &f); L->log_f(f);
	Problem::Atom s, v;

	s = P->minmax_oracle(g, x, v, NO_TIME_LIMIT); L->log_min_max_oracle();
	double gx = dot_product(x, g, n);
	double gs = P->dot_product(s, g);
	double gap = gx - gs; L->log_gap(gap);

	double alpha = P->move_DICG(xs, s, v); L->log_dicg_alpha(alpha);
	if (_s) *_s = s;
	else P->destroy_atom(s);
	P->destroy_atom(v);

	double delta_f_unconstrained;
	double gamma = P->line_search(x, xs, NULL, NULL, &delta_f_unconstrained);
	return -delta_f_unconstrained;
}

void Solver::CacheDICG_R(int depth, double T_parent, double Phi_parent, double T_limit)
{
	double W;
	double Phi = Phi_parent;
	Problem::Atom s;
	MyAtom* a;
	int t;

	double TIME_LIMIT;
	if (depth>0 && options.subproblem_time_limit>0) TIME_LIMIT = T_parent*options.subproblem_time_limit;
	else                                            TIME_LIMIT = NO_TIME_LIMIT;

	for (t=0; options.reduce_policy->reduce_allowed(depth); t++)
	{
		Problem* P0 = P;

		///////////////////////
		double T_start = _get_time();
		ComputeGapW_I(&W, &s);
//		if (Phi > W) Phi = W;
		Phi = W;
		double T_current = _get_time();
		double T = T_current - T_start;
		if ( !L->keep_going() || T_current > T_limit ) { P->destroy_atom(s); return; }
		double T_limit_R = T_limit;
		if (options.subproblem_time_limit > 0)
		{
			double TT = T_current + options.subproblem_time_limit*T;
			if (T_limit_R > TT) T_limit_R = TT;
		}
		///////////////////////

		if (depth > 0 && t > 0)
		{
			if ( Phi*Phi*T_parent <= Phi_parent*Phi_parent*T ) { P->destroy_atom(s); return; }
		}

		///////////////////////////////////////////////////////

		P = P0->reduce(g, x, s, *options.reduce_policy, TIME_LIMIT); 
		P0->destroy_atom(s);
		if (!P)
		{
			P = P0;
			return;
		}
		P0->reshape_x(NULL, P, x);
		for (a=atoms+atom_num-1; a>=atoms; a--)
		{
			a->atom = P0->reshape_atom(P0, P, a->atom);
			if (!a->atom) memcpy(a, a+1, ( (int)((atoms+(--atom_num))-a) )*sizeof(MyAtom));
		}

		n = P->get_n(); N = P->get_N(); L->log_shrink(n);
		P->init_grad(g);

		if (depth>0)
		{
			CacheDICG_R(depth+1, T_parent, Phi_parent, T_limit_R);
		}
		else
		{
			CacheDICG_R(depth+1, T, Phi, T_limit_R);
		}

		L->log_unshrink();
		n = P0->get_n(); N = P0->get_N(); 

		P0->reshape_x(P, NULL, x);
		for (a=atoms+atom_num-1; a>=atoms; a--)
		{
			a->atom = P0->reshape_atom(P, NULL, a->atom);
			if (!a->atom) memcpy(a, a+1, ( (int)((atoms+(--atom_num))-a) )*sizeof(MyAtom));
		}
		delete P;
		P = P0;
		P->init_grad(g);
	}

	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////

	int iter;
	for (iter=0; ; iter++)
	{
		double T = CacheDICGstep(Phi, NO_TIME_LIMIT);
		if (!L->keep_going()) break;
		if (_get_time()>T_limit) break;
		if (depth > 0 && Phi*Phi*T_parent < options.reduce_time_ratio*Phi_parent*Phi_parent*T) break;
	}
}

