/*
    Copyright Vladimir Kolmogorov vnk@ist.ac.at 2020-2021

    This file is part of RECURSIVE-FW.

    POTTS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "problem.h"
#include "solver.h"
#include "utils.h"

void Solver::FW()
{
	double f, gap, gamma;
	Problem::Atom s;

	P->init_grad(g);

	do
	{
		P->grad(x, g, &f); L->log_f(f);

		s = P->min_oracle(g, NO_TIME_LIMIT); L->log_min_oracle();
		P->atom_to_extended_vector(s, xs);
		P->destroy_atom(s);
		gap = dot_product_diff(x, xs, g, n); L->log_gap(gap);

		gamma = P->line_search(x, xs, NULL, NULL, NULL); L->log_line_search(gamma);
		interpolate(x, xs, gamma, x, N);
	} while (L->keep_going());

}








void Solver::DICG()
{
	P->init_grad(g);

	while ( 1 )
	{
		double W=1e300;
		DICGstep(W, NO_TIME_LIMIT);
		if (!L->keep_going()) break;
	}
}


void Solver::DICGstep(double& Phi, double time_limit)
{
	double f, alpha, gamma;
	P->grad(x, g, &f); L->log_f(f);
	Problem::Atom v;

	Problem::Atom s = P->minmax_oracle(g, x, v, time_limit); L->log_min_max_oracle();
	double gx = dot_product(x, g, n);
	double gs = P->dot_product(s, g);
	double gv = P->dot_product(v, g);

	double gap = gx - gs; L->log_gap(gap);
	double W = gv - gs; L->log_w(W);
	if (Phi > W) Phi = W;

	if (options.DICG_method == Options::PFW)
	{
		memcpy(xs, x, N*sizeof(double));
		alpha = P->move_DICG(xs, s, v); L->log_dicg_alpha(alpha);
	}
	else
	{
		if (gap >= gv - gx)
		{
			P->atom_to_extended_vector(s, xs);
			alpha = options.DICG_eps + 1.0;
		}
		else
		{
			P->atom_to_extended_vector(v, xs);
			alpha = P->move_away(x, xs); L->log_dicg_alpha(alpha);
		}
	}

	gamma = P->line_search(x, xs, NULL, NULL, NULL); L->log_line_search(gamma);
	interpolate(x, xs, gamma, x, N);

	if (alpha <= options.DICG_eps)
	{
		P->atom_to_extended_vector(s, xs);
		gamma = P->line_search(x, xs, NULL, NULL, NULL); L->log_line_search(gamma);
	}

	P->destroy_atom(s);
	P->destroy_atom(v);
}






void Solver::DICG_R()
{
	L->start_recursive(n);
	P->init_grad(g);
	DICG_R(0, 0, 1e300, _get_time() + 1e300);
}

void Solver::DICG_R(int depth, double T_parent, double Phi_parent, double T_limit)
{
	double W;
	double Phi = Phi_parent;
	Problem::Atom s;
	int t;

	for (t=0; options.reduce_policy->reduce_allowed(depth); t++)
	{
		Problem* P0 = P;

		///////////////////////
		double T_start = _get_time();
		ComputeGapW_I(&W, &s);
		//if (Phi > W) Phi = W;
		Phi = W;
		double T_current = _get_time();
		double T = T_current - T_start;
		if ( !L->keep_going() || T_current > T_limit ) { P->destroy_atom(s); return; }
		double T_limit_R = T_limit;
		if (options.subproblem_time_limit > 0)
		{
			double TT = T_current + options.subproblem_time_limit*T;
			if (T_limit_R > TT) T_limit_R = TT;
		}
		///////////////////////

		if (depth > 0 && t > 0)
		{
			if (Phi*Phi*T_parent <= Phi_parent*Phi_parent*T) { P->destroy_atom(s); return; }
		}

		///////////////////////////////////////////////////////

		P = P0->reduce(g, x, s, *options.reduce_policy, NO_TIME_LIMIT); 
		P0->destroy_atom(s);
		if (!P)
		{
			P = P0;
			break;
		}
		P0->reshape_x(NULL, P, x);

		n = P->get_n(); N = P->get_N(); L->log_shrink(n);
		P->init_grad(g);

		if (depth>0)
		{
			DICG_R(depth+1, T_parent, Phi_parent, T_limit_R);
		}
		else
		{
			DICG_R(depth+1, T, Phi, T_limit_R);
		}

		L->log_unshrink();

		P0->reshape_x(P, NULL, x);
		delete P;
		P = P0;
		n = P->get_n(); N = P->get_N(); 
		P->init_grad(g);

		if ( !L->keep_going() ) return;
	}

	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////////

	int iter;
	for (iter=0; ; iter++)
	{
		double T_start = _get_time();
//		double time_limit = options.subproblem_time_limit * Phi*Phi * T_root / (Phi_root*Phi_root);
//		double W;
		DICGstep(Phi, NO_TIME_LIMIT);
		if (!L->keep_going()) break;
		double T = _get_time();
		if (T>T_limit) break;
		if (depth > 0 && Phi*Phi*T_parent < options.reduce_time_ratio*Phi_parent*Phi_parent*(T-T_start))
		{
//printf("\n(Phi_parent,T_parent)=(%f,%f),  (Phi,T)=(%f,%f)\n", Phi_parent, T_parent, Phi, T-T_start);
			break;
		}
	}
}

double Solver::ComputeGapW_I(double* W, Problem::Atom* s_ptr)
{
	double f, gap;
	P->grad(x, g, &f); L->log_f(f);
	Problem::Atom s, v;

	if (!W)
	{
		s = P->min_oracle(g, NO_TIME_LIMIT); L->log_min_oracle();
		P->atom_to_extended_vector(s, xs);
		gap = dot_product_diff(x, xs, g, n); L->log_gap(gap);
	}
	else
	{
		s = P->minmax_oracle(g, x, v, NO_TIME_LIMIT); L->log_min_max_oracle();
		double gx = dot_product(x, g, n);
		double gs = P->dot_product(s, g);
		double gv = P->dot_product(v, g);
		P->destroy_atom(v);

		gap = gx - gs; L->log_gap(gap);
		*W = gv - gs; L->log_w(*W);
	}
	if (s_ptr) *s_ptr = s;
	else P->destroy_atom(s);
	return gap;
}

