/*
    Copyright Vladimir Kolmogorov vnk@ist.ac.at 2020-2021

    This file is part of RECURSIVE-FW.

    POTTS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __PROBLEM_H__
#define __PROBLEM_H__

#include <math.h>
#include <assert.h>
#include "timer.h"

// It is assumed that Problem uses a compact storage of atoms. 
// The user must implement a function converting atom to ExtendedVector,
// objective function f(x), and many other functions

// Vector: of size n
// ExtendedVector: of size N>=n
// For normal use, N equals N.
//
// 
// Algorithms in Solver work with vectors 'x' of size N, but use all the first 'n' components. All gradients are of size 'n'.
// However, all linear updates with vectors 'x' are performed on all N compoments.
// The user may choose to store some additional information in the last N-n components to speed up some computations (e.g. information about gradients in the case of quadratic functions).

// For an example implementation, see SIMPLEX.h

inline void PrintDoubleVector(double* x, int d)
{
	int i;
	printf("[");
	for (i=0; i<d; i++)
	{
		printf("%f", x[i]);
		if (i==d-1) printf("]");
		else        printf(", ");
	}
}

struct ReducePolicy
{
	ReducePolicy(int _max_depth=0) : depth(0), max_depth(_max_depth) {}
	virtual bool reduce_allowed(int _depth) { depth = _depth; return (depth < max_depth); }
	virtual bool is_minimal() { return (depth >= max_depth-1); }
	virtual int get_size(int n, int n_min)
	{
		int d = max_depth - depth;
		assert(d >= 2);
		double a = log((double)n), b = log((double)n_min);
		double c = a + (b-a)/d;
		int s = (int)(exp(c));
		if (s>n) s = n; if (s<n_min) s = n_min;
		return s;
	}
private:
	int depth, max_depth;
};


class Problem
{
public:
	Problem() : time_ptr(NULL) {}
	virtual ~Problem() {}

	typedef double* Vector;
	typedef double* ExtendedVector;
	typedef void* Atom;

	virtual int get_n() = 0; // size of Vector
	virtual int get_N() { return get_n(); } // size of ExtendedVector
	virtual int get_n_unique() { return get_n(); } // size of ExtendedVector

	virtual Atom get_init_atom() = 0;

	virtual void atom_to_extended_vector(Atom a, ExtendedVector x) = 0; // should convert 'a' to 'x'
	virtual void add_atom_to_extended_vector(double gamma, Atom a, ExtendedVector x) = 0; // update x += gamma*a
	virtual double dot_product(Atom a, Vector x) = 0; // return < vector(a) , x >

	virtual int compare_atoms(Atom a, Atom b) = 0; // return -1,0,1 according to some total order on atoms

	virtual double eval(ExtendedVector x) = 0; // return f(x)
	virtual void init_grad(Vector g) {} // if gradient has some constant components, then this function may set them
	virtual void grad(ExtendedVector xA, Vector g, double* f) = 0; // writes gradient to 'g', function value to '*f' (if f!=NULL)
	// return argmin_{gammain[0,1]} F(gamma)  where F(gamma) = f((1-gamma)*x + gamma*y)
	// if delta_f != NULL: let *delta_f to be the change in f (should be non-positive)
	// if delta_f1 != NULL: set *delta_f1 to  F(1)=F(0) = f(y)-f(x)
	// if delta_f_unconstrained != NULL: set *delta_f_unconstrained = min_{gamma>=0} F(gamma) - F(0)
	virtual double line_search(ExtendedVector x, ExtendedVector y, double* delta_f, double* delta_f1, double* delta_f_unconstrained) = 0;

#define NO_TIME_LIMIT 0.0
	// input: gradient g at current point x. Output: newly allocated atom 's' that minimizes < vector(s), g >
	virtual Atom min_oracle(Vector g, double time_limit) = 0;
	// input: current point x, gradient g at x. Output: newly allocated atom 's' that minimizes < vector(s), g >,
	// and atom 'v' that maximizes < vector(v), g >
	virtual Atom minmax_oracle(Vector g, Vector x, Atom& v, double time_limit)
	{
		printf("max_oracle not implemented\n");
		exit(1);
	}
	// find maximum alpha such that x + alpha*(s-v) \in P.
	// update x := x + alpha*(s-v), return alpha
	virtual double move_DICG(Vector x, Atom s, Atom v)
	{
		printf("move_DICG not implemented\n");
		exit(1);
	}

	/*
	virtual double move_towards(Vector x, Vector y)
	{
		int i, i_min = -1;
		double alpha_min=1e300;
		for (i=0; i<get_n(); i++)
		{
			double d = y[i] - x[i];
			if (d >= 0) continue;
			double alpha = -x[i]/d;
			if (alpha_min > alpha) { alpha_min = alpha; i_min = i; }
		}
		for (i=0; i<get_N(); i++)
		{
			y[i] = x[i] + alpha_min*(y[i] - x[i]);
		}
		if (i_min >= 0) y[i_min] = 0;
		return alpha_min;
	}
	*/

	// compute maximum alpha such that z = x+alpha*(x-y)\in P.
	// update y := z, return alpha
	virtual double move_away(Vector x, Vector y) 
	{
		int i, i_min = -1;
		double alpha_min=1e300;
		for (i=0; i<get_n(); i++)
		{
			if (x[i] < 0) x[i] = 0; // must be rounding error { printf("error: x[%d]=%.3e\n", i, x[i]); exit(1); }
			double d = x[i] - y[i];
			if (d >= 0) continue;
			double alpha = -x[i]/d;
			if (alpha_min > alpha) { alpha_min = alpha; i_min = i; }
		}
		for (i=0; i<get_N(); i++)
		{
			y[i] = x[i] + alpha_min*(x[i] - y[i]);
		}
		if (i_min >= 0) y[i_min] = 0;
		return alpha_min;
	}

	virtual void destroy_atom(Atom a) = 0;  // the code will eventually call this function for all atoms returned by 'get_init_atom' and 'min_oracle'

	////////////////////////////////////////////////////////////////////////
	// Input: current vector x, gradient g at x, atom 's' returned by min_oracle or minmax_oracle, 'policy'
	// output: reduced problem Q (or NULL, if no further reduction is possible)
	virtual Problem* reduce(Vector g, ExtendedVector x, Atom s, ReducePolicy& policy, double time_limit) = 0;

	// reshape atom / vector from the format of problem 'A' to the format of problem 'B'
	// A=NULL is equivalent to A=this, and similary B=NULL is equivalent to B=this
	virtual Atom reshape_atom(Problem* A, Problem* B, Atom a) = 0; // the user must deallocate 'a', and allocate new atom.
	                                                               // Note, it may happen that atom 'a' cannot be reshaped, then return NULL
	virtual void reshape_x(Problem* A, Problem* B, ExtendedVector x) = 0; // array x must have sufficient size to hold vectors both for problems 'A' and 'B'

	////////////////////////////////////////////////////////////////////////

	virtual void PrintVector(Vector x) { PrintDoubleVector((double*)x, get_n()); }
	virtual void PrintExtendedVector(ExtendedVector x) { PrintDoubleVector((double*)x, get_N()); }
	virtual void PrintAtom(Atom a) {}
	virtual void test_x_consistency(double* x) {}

	virtual void set_time_ptr(double* _time_ptr) { time_ptr = _time_ptr; }
protected:
	double* time_ptr; // if Solver::options.approx_time is true: pointer 'time_ptr' is passed to Problem P, and every operation on P must increase *time_ptr by the estimated runtime
};


#endif

