/*
    Copyright Vladimir Kolmogorov vnk@ist.ac.at 2020-2021

    This file is part of RECURSIVE-FW.

    POTTS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "problem.h"
#include "solver.h"
#include "utils.h"

Solver::Solver(Problem* _P) : P(_P), n(_P->get_n()), N(_P->get_N()), atoms(NULL), atom_num(0), L(NULL), timestamp(0)
{
	x = new double[N];
	xs = new double[N];
	g = new double[n];
	atoms = (MyAtom*) rbuf_atoms.Alloc(16*sizeof(MyAtom));
}

Solver::~Solver()
{
	delete [] x;
	delete [] xs;
	delete [] g;
	int p;
	for (p=0; p<atom_num; p++) P->destroy_atom(atoms[p].atom);
}

Solver::AtomIndex Solver::find_and_replace_atom(Problem::Atom s)
{
	AtomIndex p;
	for (p=0; p<atom_num; p++)
	{
		if (!P->compare_atoms(s, atoms[p].atom))
		{
			P->destroy_atom(atoms[p].atom);
			atoms[p].atom = s;
			return p;
		}
	}
	return -1;
}

Solver::AtomIndex Solver::add_atom(Problem::Atom s)
{
	atoms = (MyAtom*) rbuf_atoms.Realloc((atom_num + 1)*sizeof(MyAtom));
	atoms[atom_num].atom = s;
	return atom_num ++;
}

void Solver::add_atom_to_cache(Problem::Atom s)
{
	AtomIndex p;

	p = find_and_replace_atom(s);
	if (p < 0)
	{
		p = add_atom(s);
	}
	atoms[p].timestamp = timestamp;

	if (options.cache_inactive_iter_max > 0)
	{
		for (p=atom_num-1; p>=0; p--)
		{
			if (atoms[p].timestamp < timestamp - options.cache_inactive_iter_max)
			{
				remove_atom(p);
			}
		}
	}

	if (atom_num > options.cache_size_max)
	{
		AtomIndex q = 0;
		for (p=1; p<atom_num; p++)
		{
			if (atoms[q].timestamp > atoms[p].timestamp) q = p;
		}
		remove_atom(q);
	}
}


void Solver::remove_atom(AtomIndex p)
{
	P->destroy_atom(atoms[p].atom);
	memcpy(&atoms[p], &atoms[p+1], ((--atom_num) - p)*sizeof(MyAtom));
}

//////////////////////////////////////////////////////////////////////

void Solver::Solve(int iter_max, double time_max, double gap_min, const char* output_filename)
{
	Problem::Atom s = P->get_init_atom();
	P->atom_to_extended_vector(s, x);



	if (options.approx_time)
	{
		current_approx_time = 0;
		P->set_time_ptr(&current_approx_time);
	}
	else
	{
		P->set_time_ptr(NULL);
	}
	if (options.approx_time && options.subproblem_time_limit>0) { printf("ERROR: subproblem_time_limit cannot be used with approx_time\n"); exit(1); }

	L = new Logger(iter_max, time_max, gap_min, output_filename);

	AtomIndex p = add_atom(s);
	atoms[p].alpha = 1.0;

	if (options.method == Options::BCG)
	{
		BCG();
	}
	if (options.method == Options::DICG)
	{
		DICG();
	}
	if (options.method == Options::FW)
	{
		FW();
	}
	if (options.method == Options::CacheFW)
	{
		CacheFW();
	}
	if (options.method == Options::CacheDICG)
	{
		CacheDICG();
	}
	if (options.method == Options::DICG_R)
	{
		DICG_R();
	}
	if (options.method == Options::CacheDICG_R)
	{
		CacheDICG_R();
	}
	if (options.method == Options::BCG_R)
	{
		BCG_R();
	}

	double f, gap;
	P->grad(x, g, &f); L->log_f(f);
	s = P->min_oracle(g, NO_TIME_LIMIT); L->log_min_oracle();
	P->atom_to_extended_vector(s, xs);
	P->destroy_atom(s);
	gap = dot_product_diff(x, xs, g, n); L->log_gap(gap);

	delete L;

}

//////////////////////////////////////////////////////////////////////

void Solver::check_consistency()
{
	AtomIndex p;
	int i;

	double* y = new double[N];
	double alpha_sum = 0;
	memset(y, 0, N*sizeof(double));
	for (p=0; p<atom_num; p++)
	{
		alpha_sum += atoms[p].alpha;
		P->add_atom_to_extended_vector(atoms[p].alpha, atoms[p].atom, y);
	}
	double eps_max = 0;
	for (i=0; i<N; i++)
	{
		double eps = x[i] - y[i]; if (eps < 0) eps = -eps;
		if (eps_max < eps) eps_max = eps;
	}
	printf("alpha_sum=%f, eps_max=%f\n", alpha_sum, eps_max);

	delete [] y;
}

