/*
    Copyright Vladimir Kolmogorov vnk@ist.ac.at 2020-2021

    This file is part of RECURSIVE-FW.

    POTTS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __SOLVER_H__
#define __SOLVER_H__


#include "problem.h"
#include "block.h"
#include "utils.h"


class Solver
{
public:
	Solver(Problem* P);
	~Solver();

	struct Options
	{
		Options() :
			method(FW),

			reduce_policy(NULL),
			reduce_time_ratio(1.0),
			approx_time(false),
			DICG_eps(0),
			DICG_method(PFW),

			subproblem_time_limit(0),

			K1(1.0), K2(1.0), // parameters of BCG
			BCG_accept_half_step(true),

			terminate_on_simplex_failure(true),
			cache_size_max(10),
			cache_inactive_iter_max(1000000),
			cache_subiter_max(1000),
			cache_time_ratio(1.0)
		{
		}
		enum
		{
			FW, BCG, DICG, CacheFW, CacheDICG, DICG_R, BCG_R, CacheDICG_R
		} method;

		ReducePolicy* reduce_policy; // used by DICG_R, BCG_R, CacheDICG_R methods. See problems.h.
		double subproblem_time_limit; // if oracle runtime at depth 'd+1' exceeds the runtime of the parent call at depth 'd' times 'subproblem_time_limit' and subproblem_time_limit>0, then the oracle may terminate and return NULL
		double reduce_time_ratio; // multiplication for for the termination criterion of recursive calls in ..._R algorithms (default is 1, larger values lead to sooner termination)
		bool approx_time; // if false: use real clock for runtime measurements
		                  // if true:  use 'current_approx_time' instead. This variable should be updated by oracle calls, gradient computations, etc.
		                  //           (Useful to get determinstic behaviour, and might also be beneficial in a multithreaded enviroment).
		double DICG_eps; // if the size of DICG step is less than 'DICG_eps' then perform one FW step
		enum { AFW, PFW } DICG_method; // AFW = away steps, PFW = pairwise stesps
		
		///////// PARAMETERS OF BCG ///////////
		double K1, K2;
		bool BCG_accept_half_step; // if true, accept all BCG steps, including halve-steps.
		bool terminate_on_simplex_failure; // if true: if simplex step in BCG fails to improve the objective (due to rounding errors), then terminate the algorithm

		///////// PARAMETERS OF CacheDICG(_R) ///////////
		int cache_size_max; // maximum number of atoms in a cache
		int cache_inactive_iter_max; // remove atom from a cache if it is not used for more than 'cache_inactive_iter_max' iterations
		int cache_subiter_max; // maximmum number of shadow simplex steps in one iteration
		double cache_time_ratio; // multiplicative factor for the termination criterion of a given iteration (default is 1, larger values lead to sooner termination)

	} options;

	void Solve(int iter_max, double time_max, double gap_min, const char* output_filename=NULL);

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

private:
	Problem* P;
	int n, N;
	int timestamp;
	Problem::ExtendedVector x;
	Problem::ExtendedVector xs; // corresponds to s
	Problem::Vector g;

	ReusableBuffer rbuf_atoms, rbuf;

	//////////////////////////
	struct MyAtom
	{
		union
		{
			double alpha;
			int timestamp;
		};
		double c; // for temporary use

		Problem::Atom atom;
	};
	typedef int AtomIndex;
	MyAtom* atoms;
	int atom_num;

	Logger* L;

	AtomIndex find_and_replace_atom(Problem::Atom s);
	AtomIndex add_atom(Problem::Atom s);
	void remove_atom(AtomIndex);
	void add_atom_to_cache(Problem::Atom s);

	void update_weights_FW_step(double gamma, Problem::Atom s);
	void compute_c_products(Problem* Q, int& p_min, double& c_min, int& p_max, double& c_max);
	bool simplex_step(double& improvement); // assumes that MyAtom::c has been set to dot_product(atom, g). Returns true if made progress
	bool cache_step(double& improvement); // Returns true if made progress
	void BCGstep(double& Phi, bool& simplex_allowed, double time_limit);

	void FW();
	void BCG();
	void DICG();
	void CacheFW();
	void CacheDICG();
	void DICGstep(double& Phi, double time_limit);
	void CacheFWstep(double time_limit);
	double CacheDICGstep(double& Phi, double time_limit);

	void DICG_R();
	void DICG_R(int depth, double T_root, double Phi_root, double T_limit);

	void CacheDICG_R();
	void CacheDICG_R(int depth, double T_root, double Phi_root, double T_limit);

	void BCG_R();
	void BCG_R(int depth, double T_root, double Phi_root, double T_limit);

	double ComputeGapW_I(double* W=NULL, Problem::Atom* s=NULL);
	double ComputeGapW_II(double* W=NULL, Problem::Atom* s=NULL, double TIME_LIMIT=NO_TIME_LIMIT);
	double ComputeDICGimprovement(Problem::Atom* s=NULL);



	double _get_time()
	{
		return (options.approx_time) ? current_approx_time : get_time();
	}
	double current_approx_time;

	void check_consistency(); // check whether current x is indeed a convex combination of atoms
};


#endif
