/*
    Copyright Vladimir Kolmogorov vnk@ist.ac.at 2020-2021

    This file is part of RECURSIVE-FW.

    POTTS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "solver.h"

#include "PROBLEMS/PM.h"
#include "PROBLEMS/SIMPLEX.h"
#include "PROBLEMS/VIDEO_LOCALIZATION.h"
#include "PROBLEMS/DAG_SP_NODES_ONLY.h"
#include "PROBLEMS/functionQ.h"

//#define __DAG_SP_FIXED_EDGE_WEIGHTS__
//#include "PROBLEMS/DAG_SP.h"
//#undef __DAG_SP_FIXED_EDGE_WEIGHTS__

//#define __DAG_SP__
#include "PROBLEMS/DAG_SP.h"
//#undef __DAG_SP__


Problem* CreatePM()
{
	int seed = 1;
	int m = 5;
	double sigma = 2.0;
	int warm_start = false;
	bool uniform_A = false;

	PM* P = new PM(6, 6);
	P->AddEdge(0,1);
	P->AddEdge(1,2);
	P->AddEdge(2,3);
	P->AddEdge(0,3);
	P->AddEdge(3,4);
	P->AddEdge(4,5);

	P->SetRandomFunction(uniform_A, m, sigma, seed);
	P->warm_start = (warm_start) ? true : false;
	return P;
}




///////////////////////////////////////////////////////

Buffer parse_buffer(1024);

char* GetStrArg(const char* name, int& argc, char**& argv, const char* default_value=NULL)
{
	int i;
	unsigned int name_len = strlen(name);
	for (i=0; i<argc; i++)
	{
		char* s = argv[i];
		if (strlen(s) > name_len && s[name_len] == '=' && strncmp(name, s, name_len) == 0)
		{
			default_value = s+name_len+1;
			for ( ; i<argc-1; i++) argv[i] = argv[i+1];
			argc --;
			break;
		}
	}
	return (char*) default_value;
}

int GetIntArg(const char* name, int& argc, char**& argv, int default_value)
{
	char* s = GetStrArg(name, argc, argv);
	if (!s) return default_value;
	return atoi(s);
}

double GetDoubleArg(const char* name, int& argc, char**& argv, double default_value)
{
	char* s = GetStrArg(name, argc, argv);
	if (!s) return default_value;
	return atof(s);
}

// parses string of the form "name0(a1,a2,...,an)"
void ParseProblemString(char* _LINE, int& argc, char**& argv)
{
	char* LINE = (char*) parse_buffer.Alloc(strlen(_LINE)+1);
	memcpy(LINE, _LINE, strlen(_LINE)+1);
	char* c;
	bool success = false;
	for (argc=0, c=LINE; *c; c++)
	{
		if (*c == '(')
		{
			if (argc != 0) { printf("error parsing string %s\n", LINE); exit(1); }
			argc ++;
		}
		else if (*c == ',')
		{
			if (argc == 0) { printf("error parsing string %s\n", LINE); exit(1); }
			argc ++;
		}
		else if (*c == ')')
		{
			if (argc == 0 || c[1]) { printf("error parsing string %s\n", LINE); exit(1); }
			argc ++;
			success = true;
		}
	}
	if (!success) { printf("error parsing string %s\n", LINE); exit(1); }

	argv = (char**) parse_buffer.Alloc(argc*sizeof(char*));
	argv[0] = LINE;
	for (argc=0, c=LINE; *c; c++)
	{
		if (*c == '(' || *c == ',' || *c == ')')
		{
			argc ++;
			if (*c != ')') argv[argc] = c+1;
			*c = 0;
		}
	}
}

// PM(file='PROBLEMS/blossom5-instances/test.g',m=100,sigma=100)
// DAG_SP(N=100,K=5,m=100,sigma=20)
Problem* CreateProblem(char* LINE, int f_method)
{
	int argc;
	char** argv;
	ParseProblemString(LINE, argc, argv);
	char* type = argv[0];
	argv ++; argc --;
	if (!strcmp(type, "PM"))
	{
		bool uniform_A = (bool) GetIntArg("uniform_A", argc, argv, 0);
		char* filename = GetStrArg("file", argc, argv);
		int seed = GetIntArg("seed", argc, argv, -1);
		int m = GetIntArg("m", argc, argv, -1);
		double sigma = GetDoubleArg("sigma", argc, argv, -1.0);
		int warm_start = GetIntArg("warm_start", argc, argv, 0);
		if (!filename || seed < 0 || m < 0 || sigma < 0 || argc != 0) { printf("wrong problem string\n"); exit(1); }
		PM* P = new PM(filename);
		P->SetRandomFunction(uniform_A, m, sigma, seed, f_method);
		P->warm_start = (warm_start) ? true : false;
		return P;
	}
	if (!strcmp(type, "BIRKHOFF"))
	{
		bool uniform_A = (bool) GetIntArg("uniform_A", argc, argv, 0);
		int seed = GetIntArg("seed", argc, argv, -1);
		int N = GetIntArg("N", argc, argv, -1);
		int m = GetIntArg("m", argc, argv, -1);
		int degree = GetIntArg("degree", argc, argv, -1);
		double sigma = GetDoubleArg("sigma", argc, argv, -1.0);
		int warm_start = GetIntArg("warm_start", argc, argv, 0);
		if (seed < 0 || m < 0 || sigma < 0 || argc != 0) { printf("wrong problem string\n"); exit(1); }
		PM* P = new PM(2*N, N*N);
		if (degree < 0) P->SetBirkhoff();
		else P->SetRegularBipartite(degree);
		//PM* P = new PM(N);
		P->SetRandomFunction(uniform_A, m, sigma, seed, f_method);
		P->warm_start = (warm_start) ? true : false;
		return P;
	}
	if (!strcmp(type, "DAG_SP"))
	{
		bool uniform_A = (bool) GetIntArg("uniform_A", argc, argv, 0);
		int seed = GetIntArg("seed", argc, argv, -1);
		int N = GetIntArg("N", argc, argv, -1);
		int K = GetIntArg("K", argc, argv, -1);
		int m = GetIntArg("m", argc, argv, -1);
		double sigma = GetDoubleArg("sigma", argc, argv, -1.0);
		if (seed < 0 || N < 2 || K < 1 || m < 0 || sigma < 0 || argc != 0) { printf("wrong problem string\n"); exit(1); }
		DAG_SP* P = new DAG_SP(N*K, (N-1)*K*K+2*K);
		P->SetChain(N, K);
		P->SetRandomFunction(uniform_A, m, sigma, seed);
		return P;
	}
	if (!strcmp(type, "VIDEO_LOCALIZATION"))
	{
		VIDEO_LOCALIZATION* P = new VIDEO_LOCALIZATION(); //660, 2800+2400+2400+1200+2400);
		//P->Generate(seed);
		P->Load();
		return P;
	}
	if (!strcmp(type, "VIDEO_LOCALIZATION_TEST"))
	{
		int seed = GetIntArg("seed", argc, argv, -1);
		int N1 = GetIntArg("N1", argc, argv, 2);
		int K1 = GetIntArg("K1", argc, argv, 2);
		int N2 = GetIntArg("N2", argc, argv, 2);
		int K2 = GetIntArg("K2", argc, argv, 2);
		VIDEO_LOCALIZATION* P = new VIDEO_LOCALIZATION(); //660, 2800+2400+2400+1200+2400);
														  //P->Generate(seed);
		P->Generate(seed, N1, K1, N2, K2);
		return P;
	}
	if (!strcmp(type, "SIMPLEX"))
	{
		bool uniform_A = (bool) GetIntArg("uniform_A", argc, argv, 0);
		int seed = GetIntArg("seed", argc, argv, -1);
		int n = GetIntArg("n", argc, argv, -1);
		int m = GetIntArg("m", argc, argv, -1);
		int skewed = GetIntArg("skewed", argc, argv, 0);
		double sigma = GetDoubleArg("sigma", argc, argv, -1.0);
		double tau = GetDoubleArg("tau", argc, argv, 1.0);
		if (seed < 0 || n < 1 || m < 0 || sigma < 0 || argc != 0) { printf("wrong problem string\n"); exit(1); }
		SIMPLEX* P = new SIMPLEX(n);
		if (skewed) P->SetRandomFunctionSkewed(uniform_A, m, sigma, seed, f_method);
		else        P->SetRandomFunction(uniform_A, m, sigma, tau, seed, f_method);
		return P;
	}
	printf("unknown problem\n"); exit(1);
}

void Usage(char* name)
{
	printf("Usage: %s <problem>\n", name);
	exit(1);
};


void Run(int argc, char** argv)
{
	if (argc < 2) Usage(argv[0]);
	char* problem_line = argv[1];
	argc -= 2; argv += 2;
	int f_method = GetIntArg("f_method", argc, argv, 0);
	Problem* P = CreateProblem(problem_line, f_method);

	int reduce_max_depth = 0;

	Solver* S = new Solver(P);
	char* output = GetStrArg("output", argc, argv, NULL);
	int iter_max    = GetIntArg("iter_max", argc, argv, 10000000);
	double time_max = GetDoubleArg("time_max", argc, argv, 10.0);
	double gap_min  = GetDoubleArg("gap_min", argc, argv, 0.0);
	S->options.K1 = GetDoubleArg("K1", argc, argv, S->options.K1);
	S->options.K2 = GetDoubleArg("K2", argc, argv, S->options.K2);
	S->options.approx_time = GetIntArg("approx_time", argc, argv, S->options.approx_time ? 1 : 0) ? true : false;
	S->options.cache_size_max = GetIntArg("cache_size_max", argc, argv, S->options.cache_size_max);
	S->options.cache_inactive_iter_max = GetIntArg("cache_inactive_iter_max", argc, argv, S->options.cache_inactive_iter_max);
	S->options.cache_subiter_max = GetIntArg("cache_subiter_max", argc, argv, S->options.cache_subiter_max);
	S->options.cache_time_ratio = GetDoubleArg("cache_time_ratio", argc, argv, S->options.cache_time_ratio);
	S->options.reduce_time_ratio = GetDoubleArg("reduce_time_ratio", argc, argv, S->options.reduce_time_ratio);
	S->options.subproblem_time_limit = GetDoubleArg("subproblem_time_limit", argc, argv, 0.0);
	S->options.BCG_accept_half_step = (GetIntArg("BGG_accept_half_step", argc, argv, S->options.BCG_accept_half_step ? 1 : 0)) ? true : false;
	S->options.DICG_eps = GetDoubleArg("DICG_eps", argc, argv, S->options.DICG_eps);
	char* dicg_method = GetStrArg("DICG_method", argc, argv, "PFW");
	if      (!strcmp(dicg_method, "PFW"))  S->options.DICG_method = Solver::Options::PFW;
	else if (!strcmp(dicg_method, "AFW"))  S->options.DICG_method = Solver::Options::AFW;
	char* method = GetStrArg("method", argc, argv, "BCGstar");
	if      (!strcmp(method, "BCG"))  S->options.method = Solver::Options::BCG;
	else if (!strncmp(method, "BCG", 3)) { S->options.method = Solver::Options::BCG_R; reduce_max_depth = atoi(method+3); }
	else if (!strcmp(method, "DICG"))  S->options.method = Solver::Options::DICG;
	else if (!strncmp(method, "DICG", 4)) { S->options.method = Solver::Options::DICG_R; reduce_max_depth = atoi(method+4); }
	else if (!strcmp(method, "CacheFW"))  S->options.method = Solver::Options::CacheFW;
	else if (!strcmp(method, "CacheDICG"))  S->options.method = Solver::Options::CacheDICG;
	else if (!strncmp(method, "CacheDICG", 9)) { S->options.method = Solver::Options::CacheDICG_R; reduce_max_depth = atoi(method+9); }
	else if (!strcmp(method, "FW"))  S->options.method = Solver::Options::FW;
	else { printf("unknown method\n"); exit(1); }

	if (argc != 0) { printf("unknown option: %s\n", argv[0]); exit(1); }

	ReducePolicy policy(reduce_max_depth);
	S->options.reduce_policy = &policy;

	S->Solve(iter_max, time_max, gap_min, output);

	delete S;
	delete P;
}


int main(int argc, char** argv)
{
	if ( 0 )
	{
		const char* _argv[] = { "x", "BIRKHOFF(N=100,m=30,sigma=25,seed=1)", "approx_time=1", "DICG_method=AFW", "time_max=10", "iter_max=1000", "method=DICG1" };
		argv = (char**)_argv;
		argc = sizeof(_argv) / sizeof(char*);
	}

	Run(argc, argv);
}
